# This file was generated automatically by scripts/user_classes_header_generator.py do not modify it!
# If the user access files are modified, this file has to be regenerated with the script.
from typing import Optional, Union
from pandas import DataFrame


class PlotLayerHeader:
    """ 
    This class is a high level abstraction of the API, it is used to create components and data sets easily.
    """

    def activity_button(
        self,
        label: str,
        order: int,
        rows_size: Optional[int] = 1,
        cols_size: int = 2,
        align: Optional[str] = "stretch",
        padding: Optional[str] = None,
        activity_id: Optional[str] = None,
        activity_name: Optional[str] = None,
    ):
        """
        Create a button in the dashboard that executes an activity.
        :param label: the label of the button
        :param order: the order of the button
        :param rows_size: the size of the rows in the button
        :param cols_size: the size of the columns in the button
        :param align: the alignment of the button
        :param padding: the padding of the button
        :param activity_id: the id of the activity
        :param activity_name: the name of the activity
        """
        pass

    def annotated_chart(
        self,
        data: Union[list[ DataFrame], list[list[dict]]],
        order: int,
        x: str,
        y: Union[str, list[str]],
        annotations: str = "annotation",
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        title: Optional[str] = None,
        y_axis_name: Optional[str] = None,
        slider_config: Optional[dict] = None,
        slider_marks: Optional[list[tuple]] = None,
    ):
        """
        Create an annotated chart in the menu path.
        :param data: the data of the chart
        :param order: the order of the chart
        :param x: the x axis
        :param y: the y axis
        :param annotations: the annotations
        :param rows_size: the rows size of the chart
        :param cols_size: the columns size of the chart
        :param padding: the padding of the chart
        :param title: the title of the chart
        :param y_axis_name: the name of the y axis
        :param slider_config: the slider config
        :param slider_marks: the slider marks
        """
        pass

    def area(
        self,
        data: Union[str,  DataFrame, list[dict]],
        order: int,
        x: str,
        y: Union[list[str], str, None] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        title: Optional[str] = None,
        x_axis_name: Optional[str] = None,
        y_axis_name: Optional[str] = None,
        show_values: Union[list[str], str, None] = None,
        option_modifications: Optional[dict] = None,
        variant: Optional[str] = None,
    ):
        """ Create an area chart """
        pass

    def bar(
        self,
        data: Union[str,  DataFrame, list[dict]],
        order: int,
        x: str,
        y: Union[list[str], str, None] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        title: Optional[str] = None,
        x_axis_name: Optional[str] = None,
        y_axis_name: Optional[str] = None,
        show_values: Union[list[str], str, None] = None,
        option_modifications: Optional[dict] = None,
        variant: Optional[str] = None,
    ):
        """ Create a bar chart """
        pass

    def button(
        self,
        label: str,
        order: int,
        rows_size: Optional[int] = 1,
        cols_size: int = 2,
        align: Optional[str] = "stretch",
        padding: Optional[str] = None,
        on_click_events: Union[list[dict], dict, None] = None,
    ):
        """ Create a button in the dashboard. """
        pass

    def change_current_tab(
        self,
        tab: str,
    ):
        """
        Change the current tab.
        :param tab: the name of the tab
        """
        pass

    def change_path(
        self,
        path: str,
    ):
        """ Change the current path """
        pass

    def chart_and_indicators(
        self,
        order: int,
        chart_parameters: dict,
        indicators_groups: list,
        indicators_parameters: Optional[dict] = None,
        chart_rows_size: int = 3,
        cols_size: int = 12,
        chart_function: Optional[callable] = None,
    ) -> int:
        """None"""
        pass

    def chart_and_modal_button(
        self,
        order: int,
        chart_parameters: dict,
        button_modal: str,
        rows_size: int = 3,
        cols_size: int = 12,
        chart_function: Optional[callable] = None,
        button_label: str = "Read more",
        button_side_text: str = "Click on the button to read more about this topic.",
    ):
        """None"""
        pass

    def clear_context(
        self,
    ):
        """None"""
        pass

    def clear_menu_path(
        self,
    ):
        """
        Clear the current path or a subpath
        """
        pass

    def delete_chart_by_order(
        self,
        order: int,
    ):
        """
        Delete a report by order and context
        """
        pass

    def delete_modal(
        self,
        name: str,
    ):
        """ Delete a modal """
        pass

    def delete_tabs_group(
        self,
        name: str,
    ):
        """ Delete a tabs group """
        pass

    def doughnut(
        self,
        order: int,
        names: str,
        values: str,
        data: Union[str,  DataFrame, list[dict]],
        title: Optional[str] = None,
        rows_size: Optional[int] = 2,
        cols_size: Optional[int] = 3,
        padding: Optional[str] = None,
        option_modifications: Optional[dict] = None,
    ):
        """ Create a doughnut chart """
        pass

    def filter(
        self,
        order: int,
        data: str,
        field: str,
        multi_select: bool = False,
        cols_size: int = 4,
        rows_size: int = 1,
        padding: Optional[str] = None,
    ):
        """
        Filter the data set.
        :param order: the order of the chart
        :param data: the data
        :param field: the field to filter
        :param multi_select: whether to allow multi select
        :param cols_size: the size of the columns
        :param rows_size: the size of the rows
        :param padding: the padding
        """
        pass

    def free_echarts(
        self,
        data: Union[str,  DataFrame, list[dict], None] = None,
        options: Optional[dict] = None,
        raw_options: Optional[str] = None,
        order: Optional[int] = None,
        title: Optional[str] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        fields: Optional[list] = None,
        data_is_not_df: bool = False,
    ):
        """
        Create an echarts chart with custom options.
        :param data: the data of the chart
        :param options: the options of the chart
        :param raw_options: the raw options of the chart
        :param order: the order of the chart in the dashboard
        :param title: the title of the chart
        :param rows_size: the rows size of the chart
        :param cols_size: the columns size of the chart
        :param padding: the padding of the chart
        :param fields: the fields of the chart
        :param data_is_not_df: whether the data is not a dataframe
        """
        pass

    def funnel(
        self,
        order: int,
        names: str,
        values: str,
        data: Union[str,  DataFrame, list[dict]],
        title: Optional[str] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        option_modifications: Optional[dict] = None,
    ):
        """ Create a scatter chart """
        pass

    def gauge_indicator(
        self,
        order: int,
        value: int,
        title: Optional[str] = "",
        description: Optional[str] = "",
        cols_size: Optional[int] = 6,
        rows_size: Optional[int] = 1,
        color: Union[str, int] = 1,
    ):
        """ Create a gauge indicator """
        pass

    def generate_input_form_groups(
        self,
        order: int,
        form_groups: dict,
        dynamic_sequential_show: Optional[bool] = False,
        auto_send: Optional[bool] = False,
        next_group_label: Optional[str] = "Next",
        rows_size: Optional[int] = 3,
        cols_size: int = 12,
        padding: Optional[str] = None,
        modal: Optional[str] = None,
        activity_id: Optional[str] = None,
        activity_name: Optional[str] = None,
        on_submit_events: Optional[list[dict]] = None,
    ):
        """ Easier way to create an input form.
        :param menu_path: the menu path of the input form
        :param order: the order of the input form
        :param form_groups: the form groups of the input form
        :param dynamic_sequential_show: whether to show the next group after submitting the current group
        :param auto_send: whether to automatically send the form after the last group
        :param next_group_label: the label of the next group button
        :param rows_size: the rows size of the input form
        :param cols_size: the columns size of the input form
        :param padding: the padding of the input form
        :param modal: the modal to open after submitting the form
        :param activity_id: the activity id to run after submitting the form
        :param activity_name: the activity name to run after submitting the form
        :param on_submit_events: the events to run after submitting the form
        """
        pass

    def get_input_forms(
        self,
    ) -> list:
        """ Get all the input forms in the current menu_path. """
        pass

    def get_shared_data_names(
        self,
    ) -> list:
        """ Get the names of the shared data """
        pass

    def heatmap(
        self,
        order: int,
        x: str,
        y: str,
        values: str,
        data: Union[list[dict],  DataFrame, str],
        color_range: Optional[tuple[int, int]] = None,
        x_axis_name: Optional[str] = None,
        y_axis_name: Optional[str] = None,
        title: Optional[str] = None,
        padding: Optional[str] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        calculate_color_range: Optional[bool] = False,
        continuous: Optional[bool] = False,
        option_modifications: Optional[dict] = None,
        variant: Optional[str] = None,
    ):
        """ Create a heatmap chart """
        pass

    def horizontal_bar(
        self,
        data: Union[str,  DataFrame, list[dict]],
        order: int,
        x: str,
        y: Union[list[str], str, None] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        title: Optional[str] = None,
        x_axis_name: Optional[str] = None,
        y_axis_name: Optional[str] = None,
        show_values: Union[list[str], str, None] = None,
        option_modifications: Optional[dict] = None,
        variant: Optional[str] = None,
    ):
        """ Create a horizontal bar chart """
        pass

    def html(
        self,
        html: str,
        order: int,
        cols_size: Optional[int] = None,
        rows_size: Optional[int] = None,
        padding: Optional[str] = None,
    ):
        """
        Create an html report in the dashboard.
        :param html: the html code
        :param order: the order of the html
        :cols_size: the columns that the html occupies
        :rows_size: the rows that the html occupies
        :padding: padding
        """
        pass

    def iframe(
        self,
        url: str,
        order: int,
        height: int = 640,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
    ):
        """ Create an iframe report in the dashboard.
        :param url: the url of the iframe
        :param order: the order of the iframe
        :param height: the height of the iframe
        :param cols_size: the columns that the iframe occupies
        :param padding: padding
        """
        pass

    def indicator(
        self,
        data: Union[str,  DataFrame, list[dict], dict],
        order: int,
        vertical: Union[bool, str] = False,
        color_by_value: bool = False,
        cols_size: int = 12,
        rows_size: int = 1,
        padding: Optional[str] = None,
        **kwargs,
    ):
        """
        Create an indicator report in the dashboard.
        :param data: the data of the indicator
        :param order: the order of the indicator
        :param vertical: whether the indicator is vertical
        :param color_by_value: whether to color the indicator by value
        :cols_size: the columns that the indicator occupies
        :rows_size: the rows that the indicator occupies
        :padding: padding
        :title: the title of the indicator
        """
        pass

    def indicators_with_header(
        self,
        order: int,
        indicators_groups: list,
        title: str,
        indicators_parameters: Optional[dict] = None,
        subtitle: str = "",
        background_color: str = "var(--color-primary)",
        text_color: str = "var(--background-paper)",
        cols_size: int = 12,
        icon_url: str = "https://uploads-ssl.webflow.com/619f9fe98661d321dc3beec7/63e3615716d4435d29e0b82c_Acurracy.svg",
    ) -> int:
        """None"""
        pass

    def infographics_text_bubble(
        self,
        title: str,
        text: str,
        order: int,
        chart_parameters: dict,
        rows_size: int = 3,
        cols_size: int = 12,
        chart_function: Optional[callable] = None,
        background_url: Optional[str] = None,
        background_color: str = "var(--background-default)",
        bubble_location: str = "top",
        image_url: Optional[str] = None,
        image_size: int = 100,
    ):
        """None"""
        pass

    def input_form(
        self,
        options: dict,
        order: int,
        padding: Optional[str] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        modal: Optional[str] = None,
        activity_id: Optional[str] = None,
        activity_name: Optional[str] = None,
        on_submit_events: Optional[list[dict]] = None,
    ):
        """ Creates an input form.
        :param options: the options for the input form
        :param order: the order of the input form
        :param padding: the padding of the input form
        :param rows_size: the rows size of the input form
        :param cols_size: the columns size of the input form
        :param modal: the modal to open after submitting the form
        :param activity_id: the activity id to run after submitting the form
        :param activity_name: the activity name to run after submitting the form
        :param on_submit_events: the events to run after submitting the form
        """
        pass

    def line(
        self,
        data: Union[str,  DataFrame, list[dict]],
        order: int,
        x: str,
        y: Union[list[str], str, None] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        title: Optional[str] = None,
        x_axis_name: Optional[str] = None,
        y_axis_name: Optional[str] = None,
        show_values: Union[list[str], str, None] = None,
        option_modifications: Optional[dict] = None,
        variant: Optional[str] = None,
    ):
        """ Create a line chart """
        pass

    def line_and_bar_charts(
        self,
        order: int,
        x: str,
        data: Union[str,  DataFrame, list[dict], None],
        bar_names: Optional[list[str]] = None,
        line_names: Optional[list[str]] = None,
        x_axis_name: Optional[str] = None,
        bar_axis_name: Optional[str] = None,
        bar_suffix: Optional[str] = None,
        line_axis_name: Optional[str] = None,
        line_suffix: Optional[str] = None,
        title: Optional[str] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        option_modifications: Optional[dict] = None,
        variant: Optional[str] = None,
    ):
        """ Create a chart with bars and lines. """
        pass

    def line_with_confidence_area(
        self,
        data: Union[str,  DataFrame, list[dict]],
        order: int,
        x: str,
        lower: str,
        y: str,
        upper: str,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        title: Optional[str] = None,
        x_axis_name: Optional[str] = None,
        y_axis_name: Optional[str] = None,
        percentages: bool = False,
        option_modifications: Optional[dict] = None,
        variant: Optional[str] = None,
    ):
        """None"""
        pass

    def line_with_summary(
        self,
        order: int,
        data: Union[str,  DataFrame, list[dict]],
        title: str,
        x: str,
        y: Union[str, list[str], None] = None,
        description: str = "",
        value: Union[str, float, int] = "",
        rows_size: int = 2,
        cols_size: int = 2,
    ):
        """None"""
        pass

    def marked_line(
        self,
        data: Union[str,  DataFrame, list[dict]],
        order: int,
        x: str,
        marks: list,
        y: Union[list[str], str, None] = None,
        color_mark: str = "rgba(255, 173, 177, 0.4)",
        x_axis_name: Optional[str] = None,
        y_axis_name: Optional[str] = None,
        title: Optional[str] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        show_values: Union[list[str], str, None] = None,
        option_modifications: Optional[dict] = None,
        variant: Optional[str] = None,
    ):
        """ Create a predictive line chart """
        pass

    def modal_button(
        self,
        label: str,
        order: int,
        modal: str,
        rows_size: Optional[int] = 1,
        cols_size: int = 2,
        align: Optional[str] = "stretch",
        padding: Optional[str] = None,
    ):
        """
        Create a button in the dashboard that opens a modal.
        :param label: the label of the button
        :param order: the order of the button
        :param modal: the name of the modal
        :param rows_size: the size of the rows in the button
        :param cols_size: the size of the columns in the button
        :param align: the alignment of the button
        :param padding: the padding of the button
        """
        pass

    def pie(
        self,
        order: int,
        names: str,
        values: str,
        data: Union[str,  DataFrame, list[dict]],
        title: Optional[str] = None,
        rows_size: Optional[int] = 2,
        cols_size: Optional[int] = 3,
        padding: Optional[str] = None,
        option_modifications: Optional[dict] = None,
    ):
        """ Create a pie chart """
        pass

    def pop_out_of_bentobox(
        self,
    ):
        """ Stop using a bentobox """
        pass

    def pop_out_of_modal(
        self,
    ):
        """ Pop the current modal. """
        pass

    def pop_out_of_tabs_group(
        self,
    ):
        """
        Pop the current tabs index.
        """
        pass

    def predictive_line(
        self,
        data: Union[str,  DataFrame, list[dict]],
        order: int,
        x: str,
        y: Union[list[str], str, None] = None,
        min_value_mark: Union[str, int, None] = None,
        max_value_mark: Union[str, int, None] = None,
        color_mark: str = "rgba(255, 173, 177, 0.4)",
        x_axis_name: Optional[str] = None,
        y_axis_name: Optional[str] = None,
        title: Optional[str] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        show_values: Union[list[str], str, None] = None,
        option_modifications: Optional[dict] = None,
        variant: Optional[str] = None,
    ):
        """ Create a predictive line chart """
        pass

    def radar(
        self,
        order: int,
        names: str,
        data: Union[str,  DataFrame, list[dict]],
        values: Optional[list[str]] = None,
        max_field: Optional[str] = None,
        fill_area: bool = False,
        title: Optional[str] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        option_modifications: Optional[dict] = None,
    ):
        """ Create a radar chart """
        pass

    def raise_if_cant_change_path(
        self,
    ):
        """Raise an error if a tabs group or a modal is already open. """
        pass

    def rose(
        self,
        order: int,
        names: str,
        values: str,
        data: Union[str,  DataFrame, list[dict]],
        title: Optional[str] = None,
        rows_size: Optional[int] = 2,
        cols_size: Optional[int] = 3,
        padding: Optional[str] = None,
        option_modifications: Optional[dict] = None,
    ):
        """ Create a rose chart """
        pass

    def sankey(
        self,
        order: int,
        sources: str,
        targets: str,
        values: str,
        data: list,
        option_modifications: Optional[dict] = None,
        title: Optional[str] = None,
        padding: Optional[str] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
    ):
        """ Create a sankey chart """
        pass

    def scatter(
        self,
        order: int,
        point_fields: Union[list[tuple[str, str]], tuple[str, str]],
        data: Union[str,  DataFrame, list[dict]],
        x_axis_name: Optional[str] = None,
        y_axis_name: Optional[str] = None,
        title: Optional[str] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        show_values: Union[list[str], str, None] = None,
        option_modifications: Optional[dict] = None,
        variant: Optional[str] = None,
    ):
        """ Create a scatter chart """
        pass

    def scatter_with_effect(
        self,
        order: int,
        x: str,
        y: str,
        data: Union[list[str],  DataFrame, list[dict], None],
        x_axis_name: Optional[str] = None,
        y_axis_name: Optional[str] = None,
        title: Optional[str] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        effect_points: Optional[list] = None,
        option_modifications: Optional[dict] = None,
    ):
        """ Create a scatter chart """
        pass

    def segmented_area(
        self,
        data: Union[str,  DataFrame, list[dict]],
        order: int,
        x: str,
        y: str,
        segments: Optional[list] = None,
        x_axis_name: Optional[str] = None,
        y_axis_name: Optional[str] = None,
        default_color: tuple = (255, 0, 0),
        title: Optional[str] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        show_values: Union[list[str], str, None] = None,
        option_modifications: Optional[dict] = None,
        variant: Optional[str] = None,
        top_area: bool = False,
        threshold: Optional[float] = None,
        labels: Optional[list[str]] = None,
    ):
        """ Create a segmented area chart """
        pass

    def segmented_line(
        self,
        order: int,
        x: str,
        y: Union[str, list[str]],
        data: Union[str,  DataFrame, list[dict], None],
        x_axis_name: Optional[str] = None,
        y_axis_name: Optional[str] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        title: Optional[str] = None,
        marking_lines: Optional[list[int]] = None,
        range_colors: Optional[list[str]] = None,
        option_modifications: Optional[dict] = None,
        variant: Optional[str] = None,
    ):
        """ Create a chart with bars and lines. """
        pass

    def set_bentobox(
        self,
        cols_size: int,
        rows_size: int,
    ):
        """ Start using a bentobox, the id and the order will be set when the bentobox is used for the first time
        :param cols_size: the number of columns in the bentobox
        :param rows_size: the number of rows in the bentobox"""
        pass

    def set_modal(
        self,
        modal_name: str,
        width: Optional[int] = None,
        height: Optional[int] = None,
        open_by_default: Optional[bool] = None,
    ) -> None:
        """ Set the current modal.
        :param modal_name: the name of the modal
        :param width: the width of the modal
        :param height: the height of the modal
        :param open_by_default: whether the modal is open by default
        """
        pass

    def set_shared_data(
        self,
        dfs: dict = None,
        custom_data: dict = None,
    ):
        """
        Set shared data for the s.plt to use.
        :param dfs: the data frames
        :param custom_data: the custom data
        """
        pass

    def set_tabs_index(
        self,
        tabs_index: tuple,
        order: Optional[int] = None,
        parent_tabs_index: Optional[tuple[str, str]] = None,
        padding: Optional[str] = None,
        cols_size: Optional[int] = None,
        rows_size: Optional[int] = None,
        just_labels: Optional[bool] = None,
        sticky: Optional[bool] = None,
    ):
        """ Set the current tabs index.
        :param tabs_index: the index of the tabs group
        :param order: the order of the tabs group in the dashboard
        :param parent_tabs_index: the index of the parent tabs group
        :param cols_size: the size of the columns in the tabs group
        :param padding: the padding of the tabs group
        :param rows_size: the size of the rows in the tabs group
        :param just_labels: whether to show just the labels in the tabs group
        :param sticky: whether to make the tabs group sticky
        """
        pass

    def shimoku_gauge(
        self,
        order: int,
        value: Union[int, float],
        name: str = "",
        color: Union[str, int, None] = 1,
        title: Optional[str] = None,
        rows_size: Optional[int] = 1,
        cols_size: Optional[int] = 3,
        padding: Optional[str] = None,
        is_percentage: bool = False,
        option_modifications: Optional[dict] = None,
    ):
        """ Create a pie chart """
        pass

    def shimoku_gauges_group(
        self,
        gauges_data: Union[ DataFrame, list[dict]],
        order: int,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = 12,
        gauges_padding: Optional[str] = "3, 1, 1, 1",
        gauges_rows_size: Optional[int] = 9,
        gauges_cols_size: Optional[int] = 4,
        calculate_percentages: Optional[bool] = False,
    ):
        """None"""
        pass

    def single_indicator(
        self,
        data: dict,
        order: int,
        cols_size: Optional[int] = None,
        rows_size: Optional[int] = None,
        padding: Optional[str] = None,
    ):
        """
        Create an indicator report in the dashboard.
        :param data: the data of the indicator
        :param order: the order of the indicator
        :cols_size: the columns that the indicator occupies
        :rows_size: the rows that the indicator occupies
        :padding: padding
        """
        pass

    def speed_gauge(
        self,
        order: int,
        name: str,
        value: int,
        min_value: int,
        max_value: int,
        title: Optional[str] = None,
        rows_size: Optional[int] = 2,
        cols_size: Optional[int] = 3,
        padding: Optional[str] = None,
        option_modifications: Optional[dict] = None,
    ):
        """ Create a speed gauge chart """
        pass

    def stacked_area(
        self,
        data: Union[str,  DataFrame, list[dict]],
        order: int,
        x: str,
        y: Union[list[str], str, None] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        title: Optional[str] = None,
        x_axis_name: Optional[str] = None,
        y_axis_name: Optional[str] = None,
        show_values: Union[list[str], str, None] = None,
        option_modifications: Optional[dict] = None,
        variant: Optional[str] = None,
    ):
        """ Create a stacked area chart """
        pass

    def stacked_bar(
        self,
        data: Union[str,  DataFrame, list[dict]],
        order: int,
        x: str,
        y: Union[list[str], str, None] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        title: Optional[str] = None,
        x_axis_name: Optional[str] = None,
        y_axis_name: Optional[str] = None,
        show_values: Union[list[str], str, None] = None,
        option_modifications: Optional[dict] = None,
        variant: Optional[str] = None,
    ):
        """ Create a stacked bar chart """
        pass

    def stacked_horizontal_bar(
        self,
        data: Union[str,  DataFrame, list[dict]],
        order: int,
        x: str,
        y: Union[list[str], str, None] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        title: Optional[str] = None,
        x_axis_name: Optional[str] = None,
        y_axis_name: Optional[str] = None,
        show_values: Union[list[str], str, None] = None,
        option_modifications: Optional[dict] = None,
        variant: Optional[str] = None,
    ):
        """ Create a stacked horizontal bar chart """
        pass

    def sunburst(
        self,
        order: int,
        data: Union[str, dict, list],
        title: Optional[str] = None,
        padding: Optional[str] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        option_modifications: Optional[dict] = None,
    ):
        """ Create a sunburst chart """
        pass

    def table(
        self,
        order: int,
        data: Union[str,  DataFrame, list[dict], dict],
        columns: Optional[list[str]] = None,
        columns_button: bool = True,
        filters: bool = True,
        export_to_csv: bool = True,
        search: bool = True,
        page_size: int = 10,
        page_size_options: Optional[list[int]] = None,
        initial_sort_column: Optional[str] = None,
        sort_descending: bool = False,
        columns_options: Optional[dict] = None,
        categorical_columns: Optional[list[str]] = None,
        label_columns: Optional[dict] = None,
        web_link_column: Optional[str] = None,
        open_link_in_new_tab: bool = False,
        title: Optional[str] = None,
        padding: Optional[str] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
    ):
        """ Create a table report in the dashboard.
        :param order: the order of the table
        :param data: the data of the table
        :param columns: the columns of the table
        :param columns_button: whether to show the columns button
        :param filters: whether to show the filters button
        :param export_to_csv: whether to show the export to csv button
        :param search: whether to show the search bar
        :param page_size: the number of rows per page
        :param page_size_options: the options for the number of rows per page
        :param initial_sort_column: the initial sorting column
        :param sort_descending: whether to sort descending by the initial sorting column
        :param columns_options: the options for the columns
        :param categorical_columns: the categorical columns
        :param label_columns: the label columns
        :param report_params: additional report parameters as key-value pairs
        :param web_link_column: the column to use as web link
        :param open_link_in_new_tab: whether to open the web link in a new tab
        :param title: the title of the table
        :param padding: the padding of the table
        :param rows_size: the rows size of the table
        :param cols_size: the columns size of the table
        """
        pass

    def top_bottom_area(
        self,
        data: Union[str,  DataFrame, list[dict]],
        order: Optional[int] = None,
        x: str = "x",
        top_names: Optional[list[str]] = None,
        bottom_names: Optional[list[str]] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        title: Optional[str] = None,
        x_axis_name: Optional[str] = None,
        top_axis_name: Optional[str] = None,
        bottom_axis_name: Optional[str] = None,
        option_modifications: Optional[dict] = None,
        variant: Optional[str] = None,
    ):
        """None"""
        pass

    def top_bottom_line(
        self,
        data: Union[str,  DataFrame, list[dict]],
        order: int,
        x: str,
        top_names: Optional[list[str]] = None,
        bottom_names: Optional[list[str]] = None,
        rows_size: int = 4,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        title: Optional[str] = None,
        x_axis_name: Optional[str] = None,
        top_axis_name: Optional[str] = None,
        bottom_axis_name: Optional[str] = None,
        option_modifications: Optional[dict] = None,
    ):
        """None"""
        pass

    def tree(
        self,
        order: int,
        data: Union[str, dict, list[dict]],
        radial: bool = False,
        vertical: bool = False,
        title: Optional[str] = None,
        padding: Optional[str] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        option_modifications: Optional[dict] = None,
    ):
        """ Create a tree chart """
        pass

    def treemap(
        self,
        order: int,
        data: Union[dict, list, None] = None,
        title: Optional[str] = None,
        padding: Optional[str] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        option_modifications: Optional[dict] = None,
    ):
        """ Create a treemap chart """
        pass

    def waterfall(
        self,
        data: Union[str,  DataFrame, list[dict]],
        order: int,
        x: str,
        positive: str,
        negative: str,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        title: Optional[str] = None,
        x_axis_name: Optional[str] = None,
        y_axis_name: Optional[str] = None,
        show_balance: bool = False,
        variant: Optional[str] = None,
        option_modifications: Optional[dict] = None,
    ):
        """None"""
        pass

    def zero_centered_bar(
        self,
        data: Union[str,  DataFrame, list[dict]],
        order: int,
        x: str,
        y: Union[list[str], str, None] = None,
        rows_size: Optional[int] = None,
        cols_size: Optional[int] = None,
        padding: Optional[str] = None,
        title: Optional[str] = None,
        x_axis_name: Optional[str] = None,
        y_axis_name: Optional[str] = None,
        show_values: Union[list[str], str, None] = None,
        option_modifications: Optional[dict] = None,
        variant: Optional[str] = None,
    ):
        """ Create a zero centered bar chart """
        pass



import tqdm

from copy import copy
from typing import Optional, List, Dict

from shimoku import App
from shimoku.utils import create_function_name

from shimoku.code_gen.tree_generation import CodeGenTree
from shimoku.code_gen.file_generator import CodeGenFileHandler
from shimoku.code_gen.business_code_gen.apps_code_gen.\
        data_sets_code_gen.code_gen_from_data_sets import code_gen_from_shared_data_sets
from shimoku.code_gen.business_code_gen.apps_code_gen.\
        reports_code_gen.code_gen_from_reports import code_gen_from_reports_tree
from shimoku.code_gen.business_code_gen.apps_code_gen.\
        reports_code_gen.report_types_code_gen.code_gen_from_tabs import code_gen_tabs_functions
from shimoku.code_gen.business_code_gen.apps_code_gen.\
        reports_code_gen.report_types_code_gen.code_gen_from_modal import code_gen_modals_functions


import logging
from shimoku.execution_logger import ClassWithLogging
logger = logging.getLogger(__name__)


class AppCodeGen(ClassWithLogging):
    """ Class for generating code from a menu path. """

    _module_logger = logger

    def __init__(self, app: App, output_path: str, pbar: Optional[tqdm.tqdm] = None):
        self._app = app
        self.app_f_name = 'menu_path_' + create_function_name(app['name'])
        self._output_path = (f'{output_path}/'
                             f'{self.app_f_name}')
        self._actual_bentobox: Optional[Dict] = None
        self._imports_code_lines = [
            'import os',
            'from shimoku import Client',
        ]
        self._file_generator: CodeGenFileHandler = CodeGenFileHandler(self._output_path)
        self._code_gen_tree: CodeGenTree = CodeGenTree(app, self._file_generator, pbar=pbar)

    @staticmethod
    def code_gen_report_params(report: Dict) -> List[str]:
        """ Generate code for the parameters of a report.
        :param report: report to generate code from
        :return: list of code lines
        """
        report_params_to_get = {
            'order': 'order', 'title': 'title',
            'sizeColumns': 'cols_size', 'sizeRows': 'rows_size',
            'sizePadding': 'padding',
        }
        return [f'    {report_params_to_get[k]}='
                + (f'"{(report[k])}",'if isinstance(report[k], str) else f'{report[k]},')
                for k in report if k in report_params_to_get]

    async def generate_code(self):
        """ Use the resources in the API to generate code_lines for the SDK. Create a file in
        the specified path with the generated code_lines.
        """
        await self._code_gen_tree.generate_tree()
        if self._code_gen_tree.needs_pandas:
            self._imports_code_lines.extend(['import pandas as pd'])

        code_lines: List[str] = []
        function_calls_code_lines: List[str] = []

        shared_data_sets_code_lines = await code_gen_from_shared_data_sets(self)
        scripts_imports = copy(self._imports_code_lines)
        for path in self._code_gen_tree.tree:
            function_code_lines = await code_gen_from_reports_tree(self, path)
            path_name = 'sub_path_' + create_function_name(path) if path else 'no_path'
            script_code_lines = [
                *scripts_imports,
                '',
                '',
                'data_folder_path = os.path.dirname(os.path.abspath(__file__)) + "/data"',
                *await code_gen_tabs_functions(self, path),
                *await code_gen_modals_functions(self, path),
                '',
                '',
                f'def {path_name}(shimoku_client: Client):',
                *['    ' + line for line in function_code_lines],
                '',
            ]
            self._file_generator.generate_script_file(path_name, script_code_lines)
            self._imports_code_lines.append(
                f'from .{path_name} import {path_name}'
            )
            function_calls_code_lines.extend([
                '',
                f'shimoku_client.set_menu_path("{self._app["name"]}"' + (f', "{path}")' if path is not None else ')'),
                f'{path_name}(shimoku_client)'
            ])

        main_code_lines = [
            f'shimoku_client.set_menu_path("{self._app["name"]}")',
            'shimoku_client.plt.clear_menu_path()',
            *shared_data_sets_code_lines,
            *function_calls_code_lines,
        ]

        code_lines.extend([
            *self._imports_code_lines,
            '',
            '',
            'data_folder_path = os.path.dirname(os.path.abspath(__file__)) + "/data"',
            '',
            '',
            f'def {self.app_f_name}(shimoku_client: Client):',
            *['    ' + line for line in main_code_lines],
            ''
        ])

        self._file_generator.generate_script_file('app', code_lines)
        # Create an __init__.py file for the imports to work
        self._file_generator.generate_script_file('__init__', [''])



from typing import Optional

from shimoku.api.resources.app import App
from shimoku.api.resources.data_set import DataSet
from shimoku.api.resources.report import Report
from shimoku.api.resources.reports.modal import Modal
from shimoku.api.resources.reports.tabs_group import TabsGroup
from shimoku.api.utils import change_data_set_name_with_report

from shimoku.code_gen.file_generator import CodeGenFileHandler

import tqdm
import pandas as pd
import asyncio

import logging
from shimoku.execution_logger import ClassWithLogging

logger = logging.getLogger(__name__)


class CodeGenTree(ClassWithLogging):
    """ Class for generating code generation tree. """

    _module_logger = logger

    def __init__(self, app: App, file_generator: CodeGenFileHandler, pbar: Optional[tqdm.tqdm] = None):
        self._app = app
        self._file_generator = file_generator
        self._pbar = pbar
        self.custom_data_sets_with_data = {}
        self.actual_bentobox: Optional[dict] = None
        self.all_tab_groups: dict[str, list[dict]] = {}
        self.tree: Optional[dict] = None
        self.shared_data_sets: list[str] = []
        self.needs_pandas: bool = False

    def _update_pbar(self, n: int = 1):
        if self._pbar is not None:
            self._pbar.update(n)

    async def _check_for_shared_data_sets(
            self, report: Report, seen_data_sets: set, individual_data_sets: dict[str, Report]
    ):
        """ Check for shared data sets in a report.
            :param report: report to check
            :param seen_data_sets: set of data sets already seen
            :param individual_data_sets: list of data sets to append to
            """
        report_data_sets: list[Report.ReportDataSet] = await report.get_report_data_sets()
        data_sets_from_rds = set([rds['dataSetId'] for rds in report_data_sets])

        for ds_id in data_sets_from_rds:
            if ds_id in self.shared_data_sets:
                continue
            if ds_id in seen_data_sets:
                self.shared_data_sets.append(ds_id)
                del individual_data_sets[ds_id]
            else:
                individual_data_sets[ds_id] = report
                seen_data_sets.add(ds_id)

    async def _check_for_create_data_set(self, data_set: DataSet, report: Optional[Report] = None):
        """ Create a file for a data set.
        :param data_set: data set to create file for
        :param report: report to create file for
        """
        if data_set['name'] is None:
            data_set['name'] = data_set['id']

        data: list[dict] = [
            {k: v for k, v in dp.cascade_to_dict().items()
             if k not in ['id', 'dataSetId'] and v is not None}
            for dp in await data_set.get_data_points()
        ]

        if len(data) == 0:
            return

        data_as_df = pd.DataFrame(data)

        if len(data) > 1 or 'customField1' not in data[0]:
            output_name = data_set["name"] if report is None else change_data_set_name_with_report(data_set, report)
            reverse_columns = {v: k for k, v in data_set['columns'].items()} if data_set['columns'] else None
            self._file_generator.create_data_frame_file(output_name, data_as_df, reverse_columns)
            self.needs_pandas = True
        else:
            self.custom_data_sets_with_data[data_set['id']] = data[0]['customField1']

    async def _get_data_sets(self):
        """ Create files for the data sets. """
        reports = await self._app.get_reports()

        # To store the data sets in the cache for the reports to have faster access to them
        await self._app.get_data_sets()

        individual_data_sets: dict[str, Report] = {}
        seen_data_sets = set()

        await asyncio.gather(*[self._check_for_shared_data_sets(report, seen_data_sets, individual_data_sets)
                               for report in reports])
        tasks = []
        for ds_id in self.shared_data_sets:
            tasks.append(self._check_for_create_data_set(await self._app.get_data_set(ds_id)))
        for ds_id, report in individual_data_sets.items():
            tasks.append(self._check_for_create_data_set(await self._app.get_data_set(ds_id), report))
        await asyncio.gather(*tasks)

    async def _tree_from_tabs_group(
            self, tree: list, tabs_group: TabsGroup, parent_tabs_index: tuple[str, str] = None
    ):
        """ Recursively build a tree of reports from a tabs group.
        :param tree: list to append to
        :param tabs_group: tabs group to build tree from
        :param parent_tabs_index: tuple of (tabs_group hash, tab name) of parent tabs group
        """
        path = tabs_group['path'] + '_' if tabs_group['path'] else ''
        name = tabs_group['properties']['hash'][len(path):]
        tabs_group_dict = {'tabs_group': tabs_group, 'tabs': {}, 'order': tabs_group['order'],
                           'parent_tabs_index': parent_tabs_index, 'name': name}

        path = tabs_group['path']
        if path not in self.all_tab_groups:
            self.all_tab_groups[path] = []
        self.all_tab_groups[path].insert(0, tabs_group_dict)

        tree.append(tabs_group_dict)
        tabs = sorted(tabs_group['properties']['tabs'].items(), key=lambda x: x[1]['order'])
        for tab, tab_data in tabs:
            report_ids = tab_data['reportIds']
            tab_dict = {'tab_groups': [], 'other': []}
            tabs_group_dict['tabs'][tab] = tab_dict
            for child_id in report_ids:
                child_report = await self._app.get_report(child_id)
                if child_report['reportType'] == 'TABS':
                    await self._tree_from_tabs_group(
                        tab_dict['tab_groups'], child_report, (tabs_group['properties']['hash'], tab))
                else:
                    tab_dict['other'].append(child_report)
                self._update_pbar()

            tab_dict['tab_groups'] = sorted(tab_dict['tab_groups'], key=lambda x: x['tabs_group']['order'])
            tab_dict['other'] = sorted(tab_dict['other'], key=lambda x: x['order'])

    async def _tree_from_modal(self, tree: list, modal: Modal):
        """ Recursively build a tree of reports from a modal.
        :param self: PlotApi instance
        :param tree: list to append to
        :param modal: modal to build tree from
        """
        path = modal['path'] + '_' if modal['path'] else ''
        name = modal['properties']['hash'][len(path):]
        modal_dict = {'modal': modal, 'tab_groups': [], 'other': [], 'name': name}
        tree.append(modal_dict)
        for child_id in modal['properties']['reportIds']:
            child_report = await self._app.get_report(child_id)
            if child_report['reportType'] == 'TABS':
                await self._tree_from_tabs_group(modal_dict['tab_groups'], child_report)
            else:
                modal_dict['other'].append(child_report)
            self._update_pbar()

        modal_dict['tab_groups'] = sorted(modal_dict['tab_groups'], key=lambda x: x['tabs_group']['order'])
        modal_dict['other'] = sorted(modal_dict['other'], key=lambda x: x['order'])

    async def generate_tree(self):
        """ Generate a tree of reports from a list of reports.
        :param self: PlotApi instances
        """
        self.tree = {}
        reports = await self._app.get_reports()
        contained_reports = set()
        for report in reports:
            # TODO: make this not necessary
            if 'hash' not in report['properties']:
                report['properties']['hash'] = 'id_' + report['id']
            if report['reportType'] == 'MODAL':
                contained_reports.update(report['properties']['reportIds'])
            elif report['reportType'] == 'TABS':
                for tab in report['properties']['tabs']:
                    contained_reports.update(report['properties']['tabs'][tab]['reportIds'])
        reports = sorted(
            reports,
            key=lambda x: (x['pathOrder'] if x['pathOrder'] else 0,
                           '0' if x['reportType'] == 'MODAL' else
                           '1' if x['reportType'] == 'TABS' else
                           '_' + (x['properties']['hash'])),
        )
        for report in reports:

            if report['id'] in contained_reports:
                continue

            if report['path'] not in self.tree:
                self.tree[report['path']] = {'modals': [], 'tab_groups': [], 'other': []}

            if report['reportType'] == 'MODAL':
                await self._tree_from_modal(self.tree[report['path']]['modals'], report)
            elif report['reportType'] == 'TABS':
                await self._tree_from_tabs_group(self.tree[report['path']]['tab_groups'], report)
            else:
                self.tree[report['path']]['other'].append(report)

            self._update_pbar()

        for path in self.tree:
            self.tree[path]['other'] = sorted(self.tree[path]['other'], key=lambda x: x['order'])
            self.tree[path]['tab_groups'] = sorted(self.tree[path]['tab_groups'],
                                                   key=lambda x: x['tabs_group']['order'])

        # Todo: Solve path ordering
        # self.tree = {k: v for k, v in sorted(reports.items(), key=lambda item: }
        await self._get_data_sets()



from os import getenv
from shimoku.exceptions import ActivityError, CacheError
import unittest
from utils import initiate_shimoku

s = initiate_shimoku()
business_id: str = getenv('BUSINESS_ID')
mock: bool = getenv('MOCK') == 'TRUE'
async_execution: bool = getenv('ASYNC_EXECUTION') == 'TRUE'

s.set_workspace(uuid=business_id)
s.set_menu_path('test_activity')

activity_name = 'test_activity'

_activity = s.activities.get_activity(name=activity_name)
if not _activity:
    _activity = s.activities.create_activity(name=activity_name)

activity_id = _activity['id']
run_id = s.activities.create_run(uuid=activity_id)['id']


def delete_new_activity_if_it_exists():
    """
    Deletes the activity created in the test_create_delete_get_activities test.
    """
    try:
        s.activities.delete_activity(name='new_' + activity_name)
    except ActivityError:
        pass


class TestActivity(unittest.TestCase):

    def test_get_run_settings(self):
        """
        Makes a call to the activity metadata API to get run settings.
        """
        s.activities.get_run_settings(uuid=activity_id, run_id=run_id)
        s.activities.get_run_settings(uuid=activity_id, run_id=run_id)
        s.activities.get_run_settings(name=activity_name, run_id=run_id)
        s.activities.get_run_settings(name=activity_name, run_id=run_id)

    def test_create_delete_get_activities(self):
        """
        Makes various calls to the activity metadata API to create, delete and get activities in different ways.
        """

        new_activity_name = '0_new_'+activity_name
        n_activities_before = len(s.activities.get_activities())

        activity = s.activities.create_activity(name=new_activity_name)

        new_activity_id = activity['id']

        assert activity['name'] == new_activity_name
        assert activity['runs'] == []

        n_activities_mid = len(s.activities.get_activities())
        assert n_activities_mid == n_activities_before + 1

        s.set_menu_path(name='test_activity')

        n_activities_mid = len(s.activities.get_activities())
        assert n_activities_mid == n_activities_before + 1

        s.activities.delete_activity(uuid=new_activity_id)

        n_activities_after = len(s.activities.get_activities())

        assert n_activities_before == n_activities_after

        self.activity_doesnt_exist(new_activity_name)

    def test_update_activity(self):
        """
        Makes various calls to the activity metadata API to update activities.
        """
        new_activity_name = '1_new_'+activity_name

        n_activities_before = len(s.activities.get_activities())

        activity = s.activities.create_activity(name=new_activity_name)

        new_activity_id = activity['id']

        assert activity['name'] == new_activity_name

        s.activities.update_activity(uuid=new_activity_id, new_name='updated ' + new_activity_name)

        assert not s.activities.get_activity(name=new_activity_name)
        assert s.activities.get_activity(name='updated ' + new_activity_name)
        assert s.activities.create_activity(name=new_activity_name)
        self.cant_create_activity(new_activity_name)

        s.activities.delete_activity(uuid=new_activity_id)
        s.activities.delete_activity(name=new_activity_name)

        n_activities_after = len(s.activities.get_activities())

        assert n_activities_before == n_activities_after

    def test_create_get_runs(self):
        """
        Makes various calls to the activity metadata API to create, and get runs.
        """
        new_activity_name = '2_new_'+activity_name
        activity = s.activities.create_activity(name=new_activity_name)
        new_activity_id = activity['id']

        assert activity['name'] == new_activity_name
        assert activity['runs'] == []

        run1 = s.activities.create_run(name=new_activity_name)
        run2 = s.activities.create_run(uuid=new_activity_id)
        run3 = s.activities.create_run(name=new_activity_name)

        activity = s.activities.get_activity(uuid=new_activity_id, how_many_runs=1)

        assert activity['name'] == new_activity_name
        assert len(activity['runs']) == 1
        assert activity['runs'][0]['id'] in [run1['id'], run2['id'], run3['id']]

        activity = s.activities.get_activity(name=new_activity_name, how_many_runs=3)
        activities_run_ids = [run['id'] for run in activity['runs']]

        assert activity['name'] == new_activity_name
        assert len(activity['runs']) == 3
        assert run1['id'] in activities_run_ids
        assert run2['id'] in activities_run_ids
        assert run3['id'] in activities_run_ids

        # delete the activity and test that it doesn't exist
        s.activities.delete_activity(name=new_activity_name)

        self.activity_doesnt_exist(new_activity_name)

    def test_create_get_run_logs(self):
        """
        Makes various calls to the activity metadata API to create, and get logs.
        """
        new_activity_name = '3_new_'+activity_name
        activity = s.activities.create_activity(name=new_activity_name)
        new_activity_id = activity['id']

        assert activity['name'] == new_activity_name
        assert activity['runs'] == []

        run = s.activities.create_run(name=new_activity_name)

        log1 = s.activities.create_run_log(name=new_activity_name, run_id=run['id'],
                                           message='test message 1', severity='INFO')
        log2 = s.activities.create_run_log(uuid=new_activity_id, run_id=run['id'],
                                           message='test message 2', severity='DEBUG', tags={'tag1': 'tag1', 'tag2': 'tag2'})
        log3 = s.activities.create_run_log(name=new_activity_name, run_id=run['id'],
                                           message='test message 3', severity='ERROR', tags={'tag1': 'tag1'})

        activity = s.activities.get_activity(name=new_activity_name, how_many_runs=1)
        run = activity['runs'][0]
        logs = s.activities.get_run_logs(uuid=new_activity_id, run_id=run['id'])

        # test that the logs are in the correct order, they are ordered by creation time
        assert len(logs) == 3
        assert log1 == logs[0]
        assert log2 == logs[1]
        assert log3 == logs[2]

        # delete the activity and test that it doesn't exist
        s.activities.delete_activity(name=new_activity_name)

        self.activity_doesnt_exist(new_activity_name)

    def test_default_activity_settings(self):
        """
        Tests that the default settings are being set and used correctly.
        """
        new_activity_name = '4_new_' + activity_name
        # test that the default settings are being set
        settings = {'name': 'test1', 'value': 'test1'}
        activity = s.activities.create_activity(name=new_activity_name, settings=settings)

        assert activity['settings'] == settings

        # test that the default settings persist in the API
        s.set_menu_path(name='test_activity')

        activity = s.activities.get_activity(uuid=activity['id'])
        assert activity['settings'] == settings

        # delete the activity and test that it doesn't exist
        s.activities.delete_activity(name=new_activity_name)

        self.activity_doesnt_exist(new_activity_name)

    def test_get_run_logs(self):
        """
        Makes various calls to the activity metadata API to create, and get logs.
        """
        new_activity_name = '5_new_'+activity_name
        activity = s.activities.create_activity(name=new_activity_name)
        new_activity_id = activity['id']

        run = s.activities.create_run(name=new_activity_name)

        log1 = s.activities.create_run_log(name=new_activity_name, run_id=run['id'],
                                           message='test message 1', severity='INFO')
        log2 = s.activities.create_run_log(uuid=new_activity_id, run_id=run['id'],
                                           message='test message 2', severity='DEBUG', tags={'tag1': 'tag1', 'tag2': 'tag2'})
        log3 = s.activities.create_run_log(name=new_activity_name, run_id=run['id'],
                                           message='test message 3', severity='ERROR', tags={'tag1': 'tag1'})

        activity = s.activities.get_activity(name=new_activity_name, how_many_runs=1)
        run = activity['runs'][0]

        # test that the logs can be retrieved by run id
        logs = s.activities.get_run_logs(uuid=new_activity_id, run_id=run['id'])
        assert len(logs) == 3
        assert log1 == logs[0]
        assert log2 == logs[1]
        assert log3 == logs[2]

        # delete the activity and test that it doesn't exist
        s.activities.delete_activity(name=new_activity_name)

        self.activity_doesnt_exist(new_activity_name)

    def test_execute_activity(self):
        """
        Executes an activity and tests that the activity and run are created, also tests that the run has a log
        and that the settings are being passed correctly. The activity must have a webhook linked for it to work.
        """
        s.activities.create_webhook(name=activity_name, url='https://www.google.com')

        run = s.activities.execute_activity(name=activity_name)
        activity = s.activities.get_activity(name=activity_name, how_many_runs=1)

        # Runs are ordered by execution time, so the first run is the most recently executed
        assert activity['runs'][0]['id'] == run['id']
        assert len(activity['runs'][0]['logs']) == 1 or mock

        settings = [{'name': 'test1', 'value': 'test1'}, {'name': 'test2', 'value': 'test2'}]

        run1 = s.activities.execute_activity(name=activity_name, run_settings=settings[0])
        run2 = s.activities.execute_activity(name=activity_name, run_settings=settings[1])
        run3 = s.activities.create_run(name=activity_name, settings=run2['id'])
        s.activities.execute_run(name=activity_name, run_id=run3['id'])

        activity = s.activities.get_activity(name=activity_name, how_many_runs=4)

        assert activity['runs'][0]['id'] == run['id']
        assert activity['runs'][0]['settings'] == {}
        assert len(activity['runs'][0]['logs']) == 1 or mock
        assert activity['runs'][1]['id'] == run1['id']
        assert activity['runs'][1]['settings'] == settings[0]
        assert len(activity['runs'][1]['logs']) == 1 or mock
        assert activity['runs'][2]['id'] == run2['id']
        assert activity['runs'][2]['settings'] == settings[1]
        assert len(activity['runs'][2]['logs']) == 1 or mock
        assert activity['runs'][3]['id'] == run3['id']
        assert activity['runs'][3]['settings'] == settings[1]
        assert len(activity['runs'][3]['logs']) == 1 or mock

    def test_button_and_form_execute_activity(self):
        """
        Creates a button in the dashboard with the capability to execute an activity.
        """
        s.set_menu_path('test_activity')

        s.plt.activity_button(
            activity_name=activity_name, order=0, label='test_button',
            cols_size=12, align='start',
        )
        s.plt.activity_button(
            activity_name=activity_name, order=1, label='test_button',
            cols_size=12, align='center'
        )
        s.plt.activity_button(
            activity_name=activity_name, order=2, label='test_button',
            rows_size=2, cols_size=12, align='right'
        )

        for i in range(6):
            s.plt.activity_button(
                activity_name=activity_name, order=3+i, label='test_button',
                rows_size=2, cols_size=12-i*2, align='stretch', padding=f'0,{i},0,{i}'
            )

        s.plt.activity_button(
            activity_name=activity_name, order=9, label='test_button',
            cols_size=12, align='center'
        )

        form_sent = (
            "<head>"
            "<style>"  # Styles title
            ".component-title{height:auto; width:100%; "
            "border-radius:16px; padding:16px;"
            "display:flex; align-items:center;"
            "background-color:var(--chart-C1); color:var(--color-white);}"
            "<style>.base-white{color:var(--color-white);}</style>"
            "</head>"  # Styles subtitle
            "<div class='component-title'>"
            "<div class='big-icon-banner'></div>"
            "<div class='text-block'>"
            "<h1>The activity has been called</h1>"
            "<p class='base-white'>"
            "Activity Test</p>"
            "</div>"
            "</div>"
        )

        s.plt.set_modal(modal_name='Activity called')
        s.plt.html(html=form_sent, order=0)
        s.plt.pop_out_of_modal()

        form_groups = {
            'Personal information': [
                {
                    'mapping': 'name',
                    'fieldName': 'name',
                    'inputType': 'text',
                },
            ]
        }

        s.plt.generate_input_form_groups(
            order=10, form_groups=form_groups,
            activity_name=activity_name,
            modal='Activity called',
        )
        if async_execution:
            s.activate_async_execution()

    def activity_doesnt_exist(self, _activity_name):
        with self.assertRaises(ActivityError):
            s.activities.delete_activity(name=_activity_name)

    def cant_create_activity(self, _activity_name):
        with self.assertRaises(CacheError):
            s.activities.create_activity(name=_activity_name)



""""""
from os import getenv
import unittest

import pandas as pd

from shimoku.exceptions import CacheError
from shimoku import Client


api_key: str = getenv('API_TOKEN')
universe_id: str = getenv('UNIVERSE_ID')
business_id: str = getenv('BUSINESS_ID')
environment: str = getenv('ENVIRONMENT')
verbosity: str = getenv('VERBOSITY')

s = Client(
    access_token=api_key,
    universe_id=universe_id,
    environment=environment,
    verbosity='INFO',
    async_execution=True,
    retry_attempts=1
)
s.set_workspace(uuid=business_id)
s.set_menu_path('File test')
s.menu_paths.delete_all_menu_path_files(name='File test', with_shimoku_generated=True)


def test_general_files():
    data = {
        'a': [1, 2, 3],
        'b': [1, 4, 9],
    }
    df = pd.DataFrame(data)
    file_object = df.to_csv(index=False).encode('utf-8')
    filename = 'helloworld.csv'
    s.io.post_object(file_name=filename, obj=file_object)

    files = s.menu_paths.get_menu_path_files(name='File test')
    assert 'helloworld.csv' in [file['name'] for file in files]

    class MyTestCase(unittest.TestCase):
        def test_cant_create_duplicate_files(self):
            with self.assertRaises(CacheError):
                s.io.post_object(file_name=filename, obj=file_object, overwrite=False)
                s.run()

    t = MyTestCase()
    t.test_cant_create_duplicate_files()

    s.io.delete_file(file_name=filename)
    files = s.menu_paths.get_menu_path_files(name='File test')
    assert 'helloworld.csv' not in [file['name'] for file in files]


def test_tags_and_metadata():
    previous_len_files = len(s.menu_paths.get_menu_path_files(name='File test'))
    s.io.post_object(
        file_name='Shimoku generated file', obj=b'',
        tags=['shimoku_generated', 'IO'], metadata={'business_id': business_id}
    )
    len_files = len(s.menu_paths.get_menu_path_files(name='File test'))
    assert len_files == previous_len_files
    len_files = len(s.menu_paths.get_menu_path_files(name='File test', with_shimoku_generated=True))
    assert len_files == previous_len_files + 1
    file = s.io.get_object(file_name='Shimoku generated file')
    assert file == b''
    file_metadata = s.io.get_file_metadata(file_name='Shimoku generated file')
    assert file_metadata['metadata']['business_id'] == business_id
    assert file_metadata['tags'] == ['shimoku_generated', 'IO']
    s.io.delete_file(file_name='Shimoku generated file', with_shimoku_generated=True)
    len_files = len(s.menu_paths.get_menu_path_files(name='File test', with_shimoku_generated=True))
    assert len_files == previous_len_files


def test_get_object():
    file_name = 'helloworld'
    object_data = b''
    s.io.post_object(file_name=file_name, obj=object_data)
    file = s.io.get_object(file_name=file_name)
    assert file == object_data


def test_post_dataframe():
    file_name: str = 'df-test'
    df = pd.DataFrame({'a': [1, 2, 3], 'b': [1, 4, 9]})
    s.io.post_dataframe(file_name, df=df)
    file = s.io.get_dataframe(file_name=file_name)
    assert file.to_dict() == df.to_dict()


def test_post_get_model():
    from sklearn import svm
    from sklearn import datasets

    clf = svm.SVC()
    X, y = datasets.load_iris(return_X_y=True)
    clf.fit(X, y)

    s.io.post_ai_model(model_name='model-test', model=clf)
    clf2 = s.io.get_ai_model(model_name='model-test')

    assert type(clf2) == type(clf)


def test_big_data():
    filename = '../../data/bulidata.csv'
    df = pd.read_csv(filename)
    df.reset_index(inplace=True)

    s.io.post_batched_dataframe(file_name='test-big-df', df=df)

    dataset: pd.DataFrame = s.io.get_batched_dataframe(file_name='test-big-df')
    assert len(dataset) == len(df)

    s.io.post_object(file_name='test-big-df.csv', obj=df.to_csv(index=False).encode('utf-8'))


def test_more_than_100_files():
    menu_path_mt100 = 'more than 100 files test'
    s.set_menu_path(menu_path_mt100)
    s.menu_paths.delete_all_menu_path_files(name=menu_path_mt100)
    for i in range(201):
        s.io.post_object(file_name=f'file_{i}.txt', obj=b'Content of file '+str(i).encode('utf-8'))
    s.pop_out_of_menu_path()
    s.run()
    s.disable_caching()
    assert len(s.menu_paths.get_menu_path_files(name=menu_path_mt100)) == 201
    s.menu_paths.delete_all_menu_path_files(name=menu_path_mt100)
    assert len(s.menu_paths.get_menu_path_files(name=menu_path_mt100)) == 0
    s.menu_paths.delete_menu_path(name=menu_path_mt100)


test_general_files()
test_get_object()
test_post_dataframe()
test_tags_and_metadata()
test_post_get_model()
# test_big_data()
test_more_than_100_files()

s.run()



from shimoku.cli import CLIParser, CLIFuncParam
from shimoku.cli.cloud.cascade_get_resources import InitOptions, ResourceGetter
from typing import Optional
from shimoku.code_gen.main_code_gen import generate_code as gen_code
from shimoku.execution_logger import log_error
from shimoku.utils import create_function_name
from shimoku import Business
import tqdm
import tempfile
import shutil
import subprocess

import logging
logger = logging.getLogger(__name__)


def add_persistence_parser(parser: Optional[CLIParser] = None):
    """
    Function to add the playground parser to a parser
    :param parser: Parser to add the playground parser to
    """

    params = {
        'name': 'persist',
        'description': 'Commands to persist the contents of the current database',
    }

    if parser:
        persistence_parser = parser.add_command(**params)
    else:
        persistence_parser = CLIParser(**params)

    module_functions = [
        generate_code,
        commit,
        pull
    ]

    common_args = [
        CLIFuncParam(name='local-port', arg_type=int, arg_help='Local port to use', mandatory=False),
        CLIFuncParam(name='environment', arg_type=str, arg_help='Environment to use', mandatory=False),
        CLIFuncParam(name='access-token', arg_type=str, arg_help='Access token to use', mandatory=False),
        CLIFuncParam(name='universe-id', arg_type=str, arg_help='Universe ID to use', mandatory=False),
        CLIFuncParam(name='workspace-id', arg_type=str, arg_help='Workspace ID to use', mandatory=False),
    ]

    for func in module_functions:
        persistence_parser.decor_add_command(common_args=common_args)(func)

    return persistence_parser


async def generate_code(
    output_path: Optional[str],
    menu_paths: Optional[list[str]],
    hide_progress_bar: bool = CLIFuncParam(default=False, action='store_true', mandatory=False),
    use_black_formatter: bool = CLIFuncParam(default=False, action='store_true', mandatory=False),
    **kwargs
):
    """
    Generate code for a set workspace.
    :param menu_paths: List of menu paths to generate code from, leave empty to generate code for all menu paths.
    :param use_black_formatter: Use black formatter to format the generated code.
    :param output_path: Output path for the generated code.
    :param hide_progress_bar: Show progress bar while generating code.
    """
    resource_getter = ResourceGetter(InitOptions(**kwargs))
    business_object = await resource_getter.get_business()

    if not output_path:
        output_path = 'generated_code'
    pbar = None
    if not hide_progress_bar:
        pbar = tqdm.tqdm(total=sum([len(await app.get_reports())
                                    for app in await business_object.get_apps()]) + 1)
    await gen_code(
        business_object=business_object,
        output_path=output_path,
        menu_paths=menu_paths,
        use_black_formatter=use_black_formatter,
        pbar=pbar
    )
    if pbar:
        pbar.close()


async def commit(
    target_workspace: str = CLIFuncParam(prompt=True),
    target_access_token: str = CLIFuncParam(prompt=True),
    target_universe_id: str = CLIFuncParam(prompt=True),
    target_environment: str = CLIFuncParam(default='production', mandatory=False),
    hide_progress_bar: bool = CLIFuncParam(default=False, action='store_true', mandatory=False),
    **kwargs
):
    """
    Commit contents to another workspace.
    :param target_workspace: Workspace id or name to commit to.
    :param target_access_token: Access token with access to the target workspace.
    :param target_universe_id: Universe id of the target workspace.
    :param target_environment: Environment of the target workspace.
    :param hide_progress_bar: Show progress bar while generating code.
    """
    resource_getter = ResourceGetter(InitOptions(**kwargs))
    business_object = await resource_getter.get_business()
    try:
        target_resource_getter = ResourceGetter(InitOptions(
            universe_id=target_universe_id,
            access_token=target_access_token,
            environment=target_environment,
            workspace=target_workspace
        ))
        target_business_object = await target_resource_getter.get_business()
    except Exception as e:
        log_error(logger, f'Could not access target workspace: {e}', RuntimeError)
        raise e  # IDE complains if this is not here, because it thinks that log_error is not raising an exception

    temp_dir = tempfile.mkdtemp()
    pbar = None
    if not hide_progress_bar:
        pbar = tqdm.tqdm(total=sum([len(await app.get_reports())
                                    for app in await business_object.get_apps()]) + 1)
    await gen_code(
        business_object=business_object,
        business_id=target_business_object['id'],
        access_token=target_access_token,
        universe_id=target_universe_id,
        environment=target_environment,
        output_path=temp_dir,
        menu_paths=None,
        use_black_formatter=False,
        pbar=pbar
    )
    if pbar:
        pbar.close()
    process = subprocess.run(
        ['python', f'{temp_dir}/execute_workspace_{create_function_name(business_object["name"])}.py'],
        check=True
    )
    if process.returncode != 0:
        log_error(logger, 'Error while executing the generated code.', RuntimeError)
    shutil.rmtree(temp_dir)


async def pull(
    origin_workspace: str = CLIFuncParam(prompt=True),
    origin_access_token: str = CLIFuncParam(prompt=True),
    origin_universe_id: str = CLIFuncParam(prompt=True),
    origin_environment: str = CLIFuncParam(default='production', mandatory=False),
    hide_progress_bar: bool = CLIFuncParam(default=False, action='store_true', mandatory=False),
    **kwargs
):
    """
    Pull contents from another workspace.
    :param origin_workspace: Workspace id or name to pull from.
    :param origin_access_token: Access token with access to the origin workspace.
    :param origin_universe_id: Universe id of the origin workspace.
    :param origin_environment: Environment of the origin workspace.
    :param hide_progress_bar: Show progress bar while retrieving the necessary data.
    """
    resource_getter = ResourceGetter(InitOptions(**kwargs))
    business_object: Business = await resource_getter.get_business()
    try:
        origin_resource_getter = ResourceGetter(InitOptions(
            universe_id=origin_universe_id,
            access_token=origin_access_token,
            environment=origin_environment,
            workspace=origin_workspace
        ))
        origin_business_object = await origin_resource_getter.get_business()
    except Exception as e:
        log_error(logger, f'Could not access origin workspace: {e}', RuntimeError)
        raise e  # IDE complains if this is not here, because it thinks that log_error is not raising an exception

    temp_dir = tempfile.mkdtemp()
    pbar = None
    if not hide_progress_bar:
        pbar = tqdm.tqdm(total=sum([len(await app.get_reports())
                                    for app in await origin_business_object.get_apps()]) + 1)
    await gen_code(
        business_object=origin_business_object,
        business_id=business_object['id'],
        access_token=business_object.api_client.access_token,
        universe_id=business_object.parent['id'],
        environment=business_object.api_client.environment,
        output_path=temp_dir,
        menu_paths=None,
        use_black_formatter=False,
        pbar=pbar
    )
    if pbar:
        pbar.close()
    process = subprocess.run(
        ['python', f'{temp_dir}/execute_workspace_{create_function_name(origin_business_object["name"])}.py'],
        check=True
    )
    if process.returncode != 0:
        log_error(logger, 'Error while executing the generated code.', RuntimeError)
    shutil.rmtree(temp_dir)





from typing import Optional
from os import getenv
from dataclasses import dataclass
import asyncio
from shimoku.cli.utils import get_profile_config
from shimoku.exceptions import (WorkspaceError, BoardError, MenuPathError,
                                FileError, ActivityTemplateError, ActivityError,
                                RoleError)
from shimoku.execution_logger import log_error
from shimoku.api.client import ApiClient
from shimoku.api.resources.universe import Universe
from shimoku.api.resources.business import Business
from shimoku.api.resources.dashboard import Dashboard
from shimoku.api.resources.activity_template import ActivityTemplate
from shimoku.api.resources.app import App
from shimoku.api.resources.activity import Activity
from shimoku.api.resources.file import File
from shimoku.api.resources.data_set import DataSet
from shimoku.api.resources.report import Report

from shimoku.api.user_access_classes.universes_layer import UniversesLayer
from shimoku.api.user_access_classes.businesses_layer import WorkspacesLayer
from shimoku.api.user_access_classes.dashboards_layer import BoardsLayer
from shimoku.api.user_access_classes.apps_layer import MenuPathsLayer
from shimoku.api.user_access_classes.files_layer import FilesLayer
from shimoku.api.user_access_classes.activity_templates_layer import ActivityTemplatesLayer
from shimoku.api.user_access_classes.activities_layer import ActivitiesLayer
from shimoku.api.user_access_classes.data_sets_layer import DataSetsLayer
from shimoku.api.user_access_classes.reports_layer import ComponentsLayer

from shimoku.ai.ai_layer import AILayer
import logging
logger = logging.getLogger(__name__)


@dataclass
class InitOptions:

    KEY_MAP = {
        'environment': 'ENVIRONMENT',
        'access_token': 'ACCESS_TOKEN',
        'universe_id': 'UNIVERSE_ID',
        'workspace_id': 'WORKSPACE_ID',
    }

    access_token: Optional[str] = getenv('SHIMOKU_ACCESS_TOKEN')
    environment: Optional[str] = getenv('SHIMOKU_ENVIRONMENT') or 'production'
    universe_id: Optional[str] = getenv('SHIMOKU_UNIVERSE_ID')
    workspace_id: Optional[str] = getenv('SHIMOKU_WORKSPACE_ID')
    board: Optional[str] = None
    menu_path: Optional[str] = None
    role: Optional[str] = None
    activity: Optional[str] = None
    file: Optional[str] = None
    component_id: Optional[str] = None
    data_set: Optional[str] = None
    local_port: int = getenv('SHIMOKU_PLAYGROUND_PORT') or 8000

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            self.set_value(key, value)
    
    def set_value(self, name: str, value: any):
        if value:
            setattr(self, name, value)

    def __getattribute__(self, item):
        attr = object.__getattribute__(self, item)
        if attr is None and item in InitOptions.KEY_MAP:
            config = get_profile_config()
            attr = config.get(InitOptions.KEY_MAP[item])
        return attr


class ResourceGetter:

    def __init__(self, init_opts: InitOptions):
        self.init_opts = init_opts
        self.universe: Optional[Universe] = None
    
    @staticmethod
    def get_resource_by_name_or_uuid(
        resource_list: list,
        reference: str,
        name_field: str = 'name',
    ):
        return next(
            (resource
                for resource in resource_list
                if resource[name_field] == reference or resource['id'] == reference),
            None
        )

    async def get_api_client(self) -> ApiClient:
        api_client = ApiClient(
            environment=self.init_opts.environment,
            playground=self.init_opts.universe_id == 'local',
            config={'access_token': self.init_opts.access_token},
            server_port=self.init_opts.local_port,
            retry_attempts=1
        )
        api_client.semaphore = asyncio.Semaphore(api_client.semaphore_limit)  # need async for semaphore
        return api_client

    async def get_universes_layer(self) -> UniversesLayer:
        api_client: ApiClient = await self.get_api_client()
        return UniversesLayer(api_client=api_client)

    async def get_universe(self) -> Universe:
        if self.universe:
            return self.universe
        api_client: ApiClient = await self.get_api_client()
        api_client.semaphore = asyncio.Semaphore(api_client.semaphore_limit)
        return Universe(api_client, self.init_opts.universe_id)

    async def get_activity_templates_layer(self) -> ActivityTemplatesLayer:
        universe: Universe = await self.get_universe()
        return ActivityTemplatesLayer(universe=universe)

    async def get_activity_template(self) -> ActivityTemplate:
        universe: Universe = await self.get_universe()
        activity_template_ref: Optional[str] = self.init_opts.activity_template
        if not activity_template_ref:
            log_error(logger, 'Activity template not specified', ActivityTemplateError)
        activity_template: Optional[ActivityTemplate] = self.get_resource_by_name_or_uuid(
            await universe.get_activity_templates(), activity_template_ref
        )
        if not activity_template:
            log_error(logger, f'Activity template {activity_template_ref} not found', ActivityTemplateError)
        return activity_template

    @staticmethod
    def get_activity_template_fields_to_show(show_input_settings: bool) -> list[str]:
        fields = ['id', 'name', 'version']
        if show_input_settings:
            fields.append('input_settings')
        return fields

    async def get_businesses_layer(self) -> WorkspacesLayer:
        universe: Universe = await self.get_universe()
        return WorkspacesLayer(universe=universe)

    async def get_business(self) -> Business:
        universe: Universe = await self.get_universe()
        business_id = self.init_opts.workspace_id
        if not business_id:
            log_error(logger, 'Workspace not specified', WorkspaceError)
        business: Optional[Business] = await universe.get_business(business_id)
        if not business:
            log_error(logger, f'Workspace {business_id} not found', WorkspaceError)
        return business

    @staticmethod
    def get_business_fields_to_show(show_theme: bool) -> list[str]:
        fields = ['id', 'name']
        if show_theme:
            fields.append('theme')
        return fields

    async def get_dashboards_layer(self) -> BoardsLayer:
        business: Business = await self.get_business()
        return BoardsLayer(business=business)

    async def get_dashboard(self) -> Dashboard:
        business: Business = await self.get_business()
        dashboard_ref: Optional[str] = self.init_opts.board
        if not dashboard_ref:
            log_error(logger, 'Board not specified', BoardError)
        dashboard: Optional[Dashboard] = self.get_resource_by_name_or_uuid(
            await business.get_dashboards(), dashboard_ref
        )
        if not dashboard:
            log_error(logger, f'Board {dashboard_ref} not found', BoardError)
        return dashboard

    @staticmethod
    def get_dashboard_fields_to_show(show_public_permission: bool) -> list[str]:
        fields = ['id', 'name', 'order']
        if show_public_permission:
            fields.append('publicPermission')
        return fields

    async def get_apps_layer(self) -> MenuPathsLayer:
        business: Business = await self.get_business()
        return MenuPathsLayer(business=business)

    async def get_app(self, app_ref: Optional[str] = None) -> App:
        business: Business = await self.get_business()
        app_ref: Optional[str] = app_ref or self.init_opts.menu_path
        if not app_ref:
            log_error(logger, 'Menu path not specified', MenuPathError)
        app: Optional[App] = self.get_resource_by_name_or_uuid(await business.get_apps(), app_ref)
        if not app:
            log_error(logger, f'Menu path {app_ref} not found', MenuPathError)
        return app

    @staticmethod
    def get_app_fields_to_show() -> list[str]:
        return ['id', 'name', 'normalizedName', 'order',
                'hideTitle', 'hidePath', 'showBreadcrumb', 'showHistoryNavigation']

    async def get_reports_layer(self) -> ComponentsLayer:
        app: App = await self.get_app()
        return ComponentsLayer(app=app)

    async def get_report(self) -> Report:
        app: App = await self.get_app()
        return await app.get_report(uuid=self.init_opts.component_id)

    @staticmethod
    def get_report_fields_to_show(show_all_fields: bool = False) -> list[str]:
        fields = ['id', 'title', 'path', 'pathOrder', 'reportType', 'order', 'sizeColumns', 'sizeRows', 'sizePadding']
        if show_all_fields:
            fields.extend(['bentobox', 'properties', 'dataFields', 'chartData', 'smartFilters'])
        return fields

    async def get_data_sets_layer(self) -> DataSetsLayer:
        app: App = await self.get_app()
        return DataSetsLayer(app=app)

    async def get_data_set(self) -> DataSet:
        app: App = await self.get_app()
        data_set_ref: Optional[str] = self.init_opts.data_set
        if not data_set_ref:
            log_error(logger, 'Data set not specified', ActivityError)
        data_set: Optional[DataSet] = self.get_resource_by_name_or_uuid(await app.get_data_sets(), data_set_ref)
        if not data_set:
            log_error(logger, f'Data set ({data_set_ref}) not found', ActivityError)
        return data_set

    @staticmethod
    def get_data_set_fields_to_show() -> list[str]:
        return ['id', 'name']

    @staticmethod
    def get_data_fields_to_show(show_custom_field: bool) -> list[str]:
        fields = ['orderField1', 'description']
        for i in range(1, 6):
            fields.append(f'dateField{i}')
        for i in range(1, 51):
            fields.append(f'stringField{i}')
            fields.append(f'intField{i}')
        if show_custom_field:
            fields.append('customField1')
        return fields

    async def get_activities_layer(self) -> ActivitiesLayer:
        app: App = await self.get_app()
        return ActivitiesLayer(app=app)

    async def get_activity(self) -> Activity:
        app: App = await self.get_app()
        activity_ref: Optional[str] = self.init_opts.activity
        if not activity_ref:
            log_error(logger, 'Activity not specified', ActivityError)
        activity: Optional[Activity] = self.get_resource_by_name_or_uuid(await app.get_activities(), activity_ref)
        if not activity:
            log_error(logger, f'Activity {activity_ref} not found', ActivityError)
        return activity

    @staticmethod
    def get_activity_fields_to_show(show_template: bool) -> list[str]:
        fields = ['id', 'name', 'settings']
        if show_template:
            fields.extend(['activityTemplateWithMode', 'universeApiKeyId'])
        return fields

    async def get_files_layer(self) -> FilesLayer:
        app: App = await self.get_app()
        return FilesLayer(app=app)

    async def get_file(self) -> File:
        app: App = await self.get_app()
        file_ref: Optional[str] = self.init_opts.file
        if not file_ref:
            log_error(logger, 'File not specified', FileError)
        file: Optional[File] = self.get_resource_by_name_or_uuid(await app.get_files(), file_ref)
        if not file:
            log_error(logger, f'File {file_ref} not found', FileError)
        return file

    @staticmethod
    def get_file_fields_to_show() -> list[str]:
        return ['id', 'name', 'contentType', 'tags', 'metadata']

    async def get_ai_layer(self) -> AILayer:
        app = await self.get_app()
        return AILayer(
            access_token=self.init_opts.access_token,
            universe=app.parent.parent,
            app=app
        )

    async def get_role_dicts(self) -> list[dict]:
        if self.init_opts.menu_path:
            apps_layer = await self.get_apps_layer()
            app = await self.get_app()
            return await apps_layer.get_roles(app['id'])
        if self.init_opts.board:
            dashboards_layer = await self.get_dashboards_layer()
            dashboard = await self.get_dashboard()
            return await dashboards_layer.get_roles(dashboard['id'])

        business = await self.get_business()
        business_layer = await self.get_businesses_layer()
        return await business_layer.get_roles(business['id'])

    async def get_role_dict(self) -> dict:
        role_dicts = await self.get_role_dicts()
        role = self.init_opts.role
        role_obj = self.get_resource_by_name_or_uuid(role_dicts, role, 'role')
        if not role_obj:
            log_error(logger, f'Role {role} not found', RoleError)
        return role_obj



from typing import Optional, Union, TYPE_CHECKING

import pandas as pd

from shimoku.plt.EChart_definitions.default_echart_options import get_common_echart_options, get_common_series_options

if TYPE_CHECKING:
    from shimoku.plt.plt_layer import PlotLayer

from shimoku.exceptions import DataError

import logging
from shimoku.execution_logger import log_error
logger = logging.getLogger(__name__)


async def heatmap_chart(
    self: 'PlotLayer', order: int, x: str, y: str, values: str,
    data: Union[list[dict], pd.DataFrame, str], color_range: Optional[tuple[int, int]] = None,
    x_axis_name: Optional[str] = None, y_axis_name: Optional[str] = None,
    title: Optional[str] = None, padding: Optional[str] = None,
    rows_size: Optional[int] = None, cols_size: Optional[int] = None,
    calculate_color_range: Optional[bool] = False, continuous: Optional[bool] = False,
    option_modifications: Optional[dict] = None, variant: Optional[str] = None
):
    """ Create a heatmap chart """
    color_range = (0, 10) if color_range is None else color_range
    if calculate_color_range:
        aux_data = data
        if isinstance(aux_data, str):
            if aux_data not in self._shared_data:
                log_error(
                    logger,
                    f"Shared data set ({aux_data}) not found in the menu path ({str(self._app)})",
                    DataError
                )
            aux_data = self._shared_data[aux_data]
        if isinstance(aux_data, pd.DataFrame):
            aux_data = aux_data.to_dict('records')
        color_range = (min([row[values] for row in aux_data]), max([row[values] for row in aux_data]))
    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    del common_options['legend']
    del common_options['toolbox']['feature']['magicType']
    common_options['xAxis'][0].update({'type': 'category', 'data': None})
    common_options['yAxis'][0].update({'type': 'category', 'nameGap': 64})
    common_options['visualMap'] = {'calculable': True, 'orient': 'vertical',
                                   'type': 'continuous' if continuous else 'piecewise',
                                   'left': 'right', 'top': 'center', 'show': variant != 'minimal',
                                   'min': color_range[0], 'max': color_range[1]}
    common_options['grid'].update({'top': '5%', 'right': 80, 'bottom': 28, 'left': '5%' if y_axis_name else '1%'})

    series_options.update({
        'type': 'heatmap',
        'itemStyle': {'borderRadius': [0, 0, 0, 0]},
    })
    await self._create_trend_chart(
        data_mapping_to_tuples=await self._choose_data(order, data), axes=[], values=[(x, y, values)],
        echart_options=common_options, series_options=series_options, order=order, title=title, padding=padding,
        rows_size=rows_size, cols_size=cols_size, show_values='all', x_axis_names=x_axis_name, y_axis_names=y_axis_name,
        option_modifications=option_modifications, variant=variant
    )



from ..report import Report
from enum import Enum


class FilterDataSet(Report):
    """ Filter data set report class """

    report_type = 'FILTERDATASET'

    default_properties = dict(
        **Report.default_properties,
        filter=[],
        mapping=[],
    )

    class InputType(Enum):
        DATERANGE = "DATERANGE"
        NUMERIC = "NUMERIC"
        CATEGORICAL_SINGLE = "CATEGORICAL_SINGLE"
        CATEGORICAL_MULTI = "CATEGORICAL_MULTI"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



from uuid import uuid1
from typing import Optional, TYPE_CHECKING

from shimoku.api.base_resource import Resource
from shimoku.api.resources.role import Role, create_role, get_role, get_roles, delete_role
from shimoku.api.resources.event import Event
from shimoku.api.resources.dashboard import Dashboard
from shimoku.api.resources.app import App

from shimoku.utils import EventType
from shimoku.exceptions import WorkspaceError, MenuPathError, BoardError

if TYPE_CHECKING:
    from shimoku.api.resources.universe import Universe

import logging
from shimoku.execution_logger import log_error
logger = logging.getLogger(__name__)


class Business(Resource):

    _module_logger = logger
    resource_type = 'business'
    alias_field = 'name'
    plural = 'businesses'

    def __init__(self, parent: 'Universe', uuid: Optional[str] = None, alias: Optional[str] = None,
                 db_resource: Optional[dict] = None):

        params = {
            'name': alias if alias else '',
            'theme': {},
        }

        super().__init__(parent=parent, uuid=uuid, db_resource=db_resource, children=[Dashboard, App, Role, Event],
                         check_params_before_creation=['name'], params_to_serialize=['theme'],
                         params=params)

        self.currently_in_use = False

    async def delete(self):
        if self.currently_in_use:
            log_error(logger, f'Workspace {str(self)} is currently in use and cannot be deleted', WorkspaceError)
        return await self._base_resource.delete()

    async def update(self):
        return await self._base_resource.update()

    # Event methods
    async def create_event(self, event_type: EventType, content: dict, resource_id: Optional[str] = None):
        """ Creates an event but dont store it in the cache
        :param event_type: The type of the event
        :param content: The content of the event
        :param resource_id: The id of the resource that triggered the event
        """
        if event_type != EventType.BUSINESS_CONTENTS_UPDATED:
            return

        event = Event(parent=self)
        event.set_params(type=event_type.value, content=content, resourceId=resource_id)
        await event

    # Dashboard methods
    async def create_dashboard(
        self, name: str, order: int, is_public: bool = False, is_disabled: bool = False
    ) -> Dashboard:
        dashboard_metadata = dict(
            order=order,
            isDisabled=is_disabled
        )
        if is_public:
            dashboard_metadata['publicPermission'] = dict(
                isPublic=True,
                permission='READ',
                token=str(uuid1())
            )
        dashboard = await self._base_resource.create_child(Dashboard, alias=name, **dashboard_metadata)
        logger.info(f'Board {name} created with id {dashboard["id"]}')
        return dashboard

    async def update_dashboard(self, uuid: Optional[str] = None, name: Optional[str] = None, **params):
        if params.get('new_name') is not None:
            params['name'] = params.pop('new_name')
            params['new_alias'] = True
        dashboard = await self.get_dashboard(uuid, name, create_if_not_exists=False)
        if not dashboard:
            logger.warning(f'Dashboard {name} not found, cannot update it')
            return
        await self._base_resource.update_child(Dashboard, uuid=uuid, alias=name, **params)
        await self.create_event(EventType.DASHBOARD_UPDATED, params, dashboard['id'])

    async def get_dashboard(self, uuid: Optional[str] = None, name: Optional[str] = None,
                            create_if_not_exists: bool = True) -> Optional[Dashboard]:
        dashboard = await self._base_resource.get_child(Dashboard, uuid, name)

        if dashboard:
            logger.info(f'Retrieved board {dashboard["name"]} with id {dashboard["id"]}')
        elif create_if_not_exists:
            if not name:
                log_error(logger, f'Name is required to create a board', BoardError)
            how_many_dashboards = len(await self.get_dashboards())
            dashboard = await self._base_resource.create_child(Dashboard, alias=name, order=how_many_dashboards + 1)
            logger.info(f'Created board {dashboard["name"]} with id {dashboard["id"]}')
            await self.create_event(EventType.DASHBOARD_CREATED, {}, dashboard['id'])

        return dashboard

    async def get_dashboards(self) -> list[Dashboard]:
        return await self._base_resource.get_children(Dashboard)

    async def delete_dashboard(self, uuid: Optional[str] = None, name: Optional[str] = None) -> bool:
        dashboard = await self.get_dashboard(uuid, name, create_if_not_exists=False)
        result = await self._base_resource.delete_child(Dashboard, uuid, name)
        await self.create_event(EventType.DASHBOARD_DELETED, {}, dashboard['id'])
        return result

    # App methods
    async def get_app(self, uuid: Optional[str] = None, name: Optional[str] = None,
                      create_if_not_exists: bool = True) -> Optional[App]:
        app = await self._base_resource.get_child(App, uuid, name)

        if app:
            logger.info(f'Retrieved menu path {app["name"]} with id {app["id"]}')
        elif create_if_not_exists:
            if not name:
                log_error(logger, f'Name is required to create a menu path', MenuPathError)
            how_many_apps = len(await self.get_apps())
            app = await self._base_resource.create_child(App, alias=name, order=how_many_apps + 1)

            logger.info(f'Created menu path {app["name"]} with id {app["id"]}')
            await self.create_event(EventType.APP_CREATED, {}, app['id'])

        return app

    async def get_apps(self) -> list[App]:
        return await self._base_resource.get_children(App)

    async def delete_app(self, uuid: Optional[str] = None, name: Optional[str] = None) -> bool:
        app = await self.get_app(uuid, name, create_if_not_exists=False)
        result = await self._base_resource.delete_child(App, uuid, name)
        await self.create_event(EventType.APP_DELETED, {}, app['id'])
        if self.api_client.playground:
            dashboards = await self.get_dashboards()
            for dashboard in dashboards:
                if app['id'] in await dashboard.list_app_ids():
                    await dashboard.remove_app(app)
        return result

    async def update_app(self, uuid: Optional[str] = None, name: Optional[str] = None, **params):
        if params.get('new_name') is not None:
            params['name'] = params.pop('new_name')
            params['new_alias'] = True
        app = await self.get_app(uuid, name, create_if_not_exists=False)
        if not app:
            logger.warning(f'App {name} not found, cannot update it')
            return
        await self._base_resource.update_child(App, uuid=uuid, alias=name, **params)
        await self.create_event(EventType.APP_UPDATED, params, app['id'])

    # Role methods
    get_role = get_role
    get_roles = get_roles
    create_role = create_role
    delete_role = delete_role



import os
import json
from os import getenv
from typing import Optional
from rich.console import Console
from rich.table import Table
from rich import print
import prompt_toolkit
from shimoku import Client

import logging

home_directory = os.path.expanduser('~')
SHIMOKU_PATH = os.path.join(home_directory, '.shimoku')
if not os.path.exists(SHIMOKU_PATH):
    os.makedirs(SHIMOKU_PATH)

CONFIG_PATH = os.path.join(SHIMOKU_PATH, 'config.json')
if not os.path.exists(CONFIG_PATH):
    with open(CONFIG_PATH, 'w') as f:
        f.write('')

INTERACTIVE_HISTORY_PATH = os.path.join(SHIMOKU_PATH, 'interactive_history.txt')
if not os.path.exists(INTERACTIVE_HISTORY_PATH):
    with open(INTERACTIVE_HISTORY_PATH, 'w') as f:
        f.write('')


environment_v = 'ENVIRONMENT'
access_token_v = 'ACCESS_TOKEN'
universe_id_v = 'UNIVERSE_ID'
workspace_id_v = 'WORKSPACE_ID'


def get_config_profiles() -> list[str]:
    """ Function to get the configuration profiles
    :return: List of configuration profiles
    """
    config_file = open(CONFIG_PATH, 'r')
    try:
        all_configs = json.loads(config_file.read()).get('profiles', {})
        profiles = list(all_configs.keys())
    except json.decoder.JSONDecodeError:
        profiles = []
    config_file.close()
    return profiles


def get_current_profile() -> str:
    """ Function to get the current configuration profile
    :return: Current configuration profile
    """
    config_file = open(CONFIG_PATH, 'r')
    try:
        all_configs = json.loads(config_file.read())
        current_profile = all_configs.get('current_profile', 'default')
    except json.decoder.JSONDecodeError:
        current_profile = 'default'
    config_file.close()
    return current_profile


def get_profile_config(profile: Optional[str] = None) -> dict:
    """ Function to get the configuration from the config file
    :return: Configuration dictionary
    """
    config_file = open(CONFIG_PATH, 'r')
    try:
        all_configs = json.loads(config_file.read())
        current_profile = profile or all_configs.get('current_profile', 'default')
        if 'profiles' not in all_configs:
            raise json.decoder.JSONDecodeError
        for config in all_configs['profiles'].values():
            if any(key not in config for key in [environment_v, access_token_v, universe_id_v, workspace_id_v]):
                raise json.decoder.JSONDecodeError
        config = all_configs['profiles'][current_profile]
    except json.decoder.JSONDecodeError:
        config = {
            environment_v: 'production',
            access_token_v: 'local',
            universe_id_v: 'local',
            workspace_id_v: 'local',
        }
    config_file.close()
    return config


def get_all_configs() -> dict:
    """ Function to get all the configurations from the config file
    :return: Configuration dictionary
    """
    config_file = open(CONFIG_PATH, 'r')
    try:
        all_configs = json.loads(config_file.read())
    except json.decoder.JSONDecodeError:
        all_configs = {}
    config_file.close()
    return all_configs


def filter_dict_list(
        l: list[dict], field: str,
        prefix: Optional[str] = None,
        contains: Optional[str] = None,
        case_sensitive: Optional[bool] = False,
        sort_field: Optional[str] = None,
        equal_to: Optional[str] = None,
        not_equal_to: Optional[str] = None,
) -> list[dict]:
    """ Function to filter a list of dictionaries
    :param l: List of dictionaries
    :param field: Field to filter
    :param prefix: Prefix to filter
    :param contains: String to filter
    :param case_sensitive: Flag to set the case sensitivity of the contains filter
    :param sort_field: Field to sort the list
    :param equal_to: String to match the items
    :param not_equal_to: String to not match the items
    :return: Filtered list of dictionaries
    """
    if not sort_field:
        sort_field = field
    if not case_sensitive:
        prefix = prefix.lower() if prefix else None
        contains = contains.lower() if contains else None
    filtered_list = []
    for element in l:
        if not isinstance(element[field], dict) and not isinstance(element[field], list):
            value = str(element[field])
        else:
            value = json.dumps(element[field], indent=2)
        if equal_to and value != equal_to:
            continue
        if not_equal_to and value == not_equal_to:
            continue
        value = value if case_sensitive else value.lower()
        if prefix and not value.startswith(prefix):
            continue
        if contains and contains not in value:
            continue
        filtered_list.append(element)

    try:
        non_none_items = [x[sort_field] for x in filtered_list if x[sort_field] is not None]
        if len(non_none_items) == 0:
            print(f"Can't sort the list by {sort_field} because all the items are None")
            return filtered_list
        list_type = type(non_none_items[0])
        for item in filtered_list:
            if item[sort_field] is None:
                item[sort_field] = list_type()
        return sorted(
            filtered_list,
            key=lambda x: x[sort_field]
            if not isinstance(x[sort_field], dict) else str(x[sort_field])
        )
    except TypeError:
        print(f"Error sorting the list by {sort_field}")
        return filtered_list


def display_dict_list(
        items: list[dict], title: str,
        fields: Optional[list[str]] = None,
        table_lines: bool = False,
):
    """ Function to display a list of dictionaries
    :param items: List of dictionaries
    :param title: Title of the table
    :param fields: Fields to display
    :param table_lines: Flag to set the table lines
    """
    if fields is None:
        fields = list(items[0].keys())
    table = Table(title=title, show_lines=table_lines)
    for field in fields:
        table.add_column(field)
    for item in items:
        row = []
        for field in fields:
            if not isinstance(item[field], dict) and not isinstance(item[field], list):
                row.append(str(item[field]))
            else:
                row.append(json.dumps(item[field], indent=2))
        table.add_row(*row)

    console = Console()
    print()
    console.print(table)
    print()


def list_filtering_and_display(
        items: list[dict],
        table_title: str,
        fields: Optional[list[str]] = None,
        table_lines: bool = True,
        **kwargs,
):
    """ Function to filter and display a list of items
    :param items: List of items to filter and display
    :param table_title: Title of the table
    :param fields: Fields to display
    :param table_lines: Flag to set the table lines
    """
    if not items:
        print("\nNo items to list\n")
        return

    if fields is None:
        fields = list(items[0].keys())
    else:
        fields = [field for field in fields if field in items[0].keys()]

    if not fields:
        print("\nNo fields to display\n")
        return

    if not kwargs['filter_field']:
        kwargs['filter_field'] = fields[0]
    if not kwargs['sort_field']:
        kwargs['sort_field'] = kwargs['filter_field']

    items = filter_dict_list(
        items,
        field=kwargs['filter_field'],
        prefix=kwargs['prefix'],
        contains=kwargs['contains'],
        case_sensitive=kwargs['case_sensitive'],
        sort_field=kwargs['sort_field'],
        equal_to=kwargs['equal_to'],
        not_equal_to=kwargs['not_equal_to'],
    )

    if kwargs['count']:
        print(f"# of {table_title}: {len(items)}")
        return
    else:
        display_dict_list(
            items,
            title=table_title,
            fields=fields,
            table_lines=table_lines,
        )


def display_dict(
        item: dict,
        fields: list[str]
):
    """ Function to display a dictionary
    :param item: Dictionary to display
    :param fields: Fields to display
    """
    resource_dict = {k: v for k, v in item.items() if k in fields}
    print(json.dumps(resource_dict, indent=4))


def display_list(
    title: str,
    items: list[str],
):
    """ Function to display a list
    :param title: Title of the list
    :param items: List to display
    """
    if not items:
        print("\nNo items to list\n")
        return
    print(f"\n{title}:")
    for item in items:
        print('  · ' + item)
    print()


def choose_from_menu(options: list[str], title: str = 'Choose an option: ') -> str:
    """
    Function to choose from a list of options
    :param options: List of options to choose from
    :param title: Title of the menu
    """
    completer = prompt_toolkit.completion.WordCompleter(options)
    selected = prompt_toolkit.prompt(title, completer=completer)
    while selected not in options:
        print("Invalid option, use the TAB key to see the available options")
        selected = prompt_toolkit.prompt(title, completer=completer)
    return selected


def input_list() -> list[str]:
    """ Function to input a list of strings
    :return: List of strings
    """
    print("Enter the list of strings, one per line, press Enter twice to finish")
    items = []
    while True:
        item = input()
        if item:
            items.append(item)
        else:
            break
    return items


def input_dict() -> dict[str, str]:
    """ Function to input a dictionary of strings
    :return: Dictionary of strings
    """
    print("Enter the dictionary of strings, first the key and then the value, press Enter in the key step to finish")
    items = {}
    while True:
        key = input("Key: ")
        if key:
            value = input("Value: ")
            items[key] = value
        else:
            break
    return items


def save_as_file(
    logger: logging.Logger,
    file_path: str,
    file_obj: bytes,
    file_name: str = 'output.txt',
):
    """ Function to save content to a file
    :param logger: Logger to use
    :param file_path: Path to save the file
    :param file_obj: File object to save
    :param file_name: Name of the file to save
    """
    if len(file_name.split('.')) == 1:
        file_name += '.txt'
    try:
        if not os.path.exists(file_path):
            os.makedirs(file_path)
        file_path = os.path.join(file_path, file_name)
        with open(file_path, 'wb') as write_file:
            write_file.write(file_obj)
            logger.info(f'File saved to {file_path}')
    except Exception as e:
        logger.error(f'Error saving file to {file_path}: {e}')
        return



from typing import Optional, List, Tuple

from shimoku.api.base_resource import Resource
from shimoku.api.resources.business import Business
from shimoku.api.resources.activity_template import ActivityTemplate
from shimoku.api.client import ApiClient

import logging
from shimoku.execution_logger import log_error
logger = logging.getLogger(__name__)


class Universe(Resource):

    _module_logger = logger
    resource_type = 'universe'

    def __init__(self, api_client: ApiClient, uuid: str):
        super().__init__(api_client=api_client, uuid=uuid, children=[Business, ActivityTemplate])

    async def create_universe_api_key(
        self, description: str
    ) -> dict:
        endpoint = self._base_resource.base_url + f'universe/{self._base_resource.id}/apiKey'

        params = dict(
            userType='ADMIN',
            enabled=True,
            description=description
        )

        return await self._base_resource.api_client.query_element(
            method='POST', endpoint=endpoint,
            **{'body_params': params}
        )

    async def get_universe_api_keys(self) -> list[dict]:
        endpoint = self._base_resource.base_url + f'universe/{self._base_resource.id}/apiKeys'

        return (await self._base_resource.api_client.query_element(
            method='GET', endpoint=endpoint,
        ))['items']

    async def delete_universe_api_key(self, uuid: str) -> bool:
        endpoint = self._base_resource.base_url + f'universe/{self._base_resource.id}/apiKey/{uuid}'

        return await self._base_resource.api_client.query_element(
            method='DELETE', endpoint=endpoint,
        )

    # Business methods
    async def create_business(
        self, name: str, theme: Optional[dict] = None
    ) -> Business:
        if self._base_resource.api_client.playground:
            log_error(logger, 'Cannot create business in local environment', RuntimeError)
        return await self._base_resource.create_child(Business, alias=name, theme=theme if theme else {})

    async def update_business(
        self, uuid: Optional[str] = None, name: Optional[str] = None, **params
    ) -> bool:
        if params.get('new_name') is not None:
            params['name'] = params.pop('new_name')
            params['new_alias'] = True
        if self._base_resource.api_client.playground:
            uuid, name = 'local', None
        return await self._base_resource.update_child(Business, uuid=uuid, alias=name, **params)

    async def get_business(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ) -> Optional[Business]:
        if self._base_resource.api_client.playground:
            uuid, name = 'local', None
        return await self._base_resource.get_child(Business, uuid, name)

    async def get_businesses(self) -> List[Business]:
        return await self._base_resource.get_children(Business)

    async def delete_business(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ) -> bool:
        if self._base_resource.api_client.playground:
            log_error(logger, 'Cannot delete local business', RuntimeError)
        return await self._base_resource.delete_child(Business, uuid, name)

    # Activity template methods
    async def get_activity_template(
            self, uuid: Optional[str] = None, name_version: Optional[Tuple[str, str]] = None
    ) -> Optional[ActivityTemplate]:
        return await self._base_resource.get_child(ActivityTemplate, uuid, name_version)

    @staticmethod
    def interpret_version(
        version: str
    ) -> list:
        """ Interpret a version string by appending '0' to each section for sorting. """
        numbers = '0123456789'
        if not version:
            return [0, 0, 0]

        def interpret_version_section(section: str) -> list:
            """ Interpret a version section, considering non-numeric characters as earlier. """
            res_section = []
            number = ''
            for c in section:
                if c in numbers:
                    number += c
                else:
                    if number:
                        res_section.append(ord('z') + int(number))
                        number = ''
                    res_section.append(ord(c))
            if number:
                res_section.append(ord('z') + int(number))
            return res_section

        return [interpret_version_section(section + '0') for section in version.split('.')]

    async def get_activity_templates(self) -> list[ActivityTemplate]:
        templates = await self._base_resource.get_children(ActivityTemplate)
        return sorted(templates, key=lambda t: [t['name']] + self.interpret_version(t['version']))



# This file was generated automatically by scripts/user_classes_header_generator.py do not modify it!
# If the user access files are modified, this file has to be regenerated with the script.
from typing import Optional, Union
from pandas import DataFrame


class DataSetsLayerHeader:
    """
    This class is used to interact with the API at the data set level.
    """

    def append_to_data_set(
        self,
        data: Union[list[dict],  DataFrame],
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> str:
        """
        Appends data to a dataset, if the dataset does not exist it will be created.
        :param name: name of the dataset
        :param uuid: uuid of the dataset
        :param data: data to be stored in the dataset
        :return: dataset id
        """
        pass

    def delete_data_set(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """
        Delete a dataset in the app
        :param name: name of the dataset
        :param uuid: uuid of the dataset
        """
        pass

    def get_data_from_data_set(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        limit: Optional[int] = None,
    ) -> list:
        """ Get the data in a dataset in the app
        :param name: name of the dataset
        :param uuid: uuid of the dataset
        :param limit: limit of the data points to be returned
        :return: data in the dataset
        """
        pass

    def replace_data_from_data_set(
        self,
        data: Union[list[dict],  DataFrame],
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """ Replace the data in a dataset in the app
        :param name: name of the dataset
        :param uuid: uuid of the dataset
        :param data: data to be stored in the dataset
        """
        pass



from typing import Optional, Union, TYPE_CHECKING
from math import floor, ceil, log10
from pandas import DataFrame
from copy import deepcopy

from shimoku.plt.utils import validate_data_is_pandarable, ShimokuPalette
from shimoku.plt.EChart_definitions.default_echart_options import get_common_echart_options, get_common_series_options

if TYPE_CHECKING:
    from shimoku.plt.plt_layer import PlotLayer


async def line_and_bar_charts(
    self: 'PlotLayer', order: int, x: str,
    data: Optional[Union[str, DataFrame, list[dict]]],
    bar_names: Optional[list[str]] = None,
    line_names: Optional[list[str]] = None,
    x_axis_name: Optional[str] = None,
    bar_axis_name: Optional[str] = None, bar_suffix: Optional[str] = None,
    line_axis_name: Optional[str] = None, line_suffix: Optional[str] = None,
    title: Optional[str] = None, rows_size: Optional[int] = None, cols_size: Optional[int] = None,
    padding: Optional[str] = None,
    option_modifications: Optional[dict] = None,
    variant: Optional[str] = None
):
    """ Create a chart with bars and lines. """
    data_mappings_to_tuples = await self._choose_data(order, data)
    if isinstance(data, str):
        data = self._shared_data[data]

    if bar_names is None:
        bar_names = ['bar']

    if line_names is None:
        line_names = ['line']

    df: DataFrame = validate_data_is_pandarable(data)
    max_bars = df[bar_names].max().max()
    pow_10 = floor(log10(abs(max_bars)))
    max_bars = ceil(max_bars / 10 ** pow_10) * 10 ** pow_10
    bar_interval = abs(max_bars) / (5 if 1000 > max_bars > 0.01 else 10)

    longest_value_bars = max(len(str(max_bars)), len(str(bar_interval)))

    max_lines = df[line_names].max().max()
    pow_10 = floor(log10(abs(max_lines)))
    max_lines = ceil(max_lines / 10 ** pow_10) * 10 ** pow_10
    line_interval = abs(max_lines) / (5 if 1000 > max_lines > 0.01 else 10)

    longest_value_lines = max(len(str(max_lines)), len(str(line_interval)))

    common_options = get_common_echart_options()
    x_axis = common_options['xAxis'][0]
    del x_axis['data']
    x_axis['type'] = 'value'
    x_axis['scale'] = True
    common_options['yAxis'][0]['scale'] = True

    common_options.update({
        'xAxis': [{
            'data': '#set_data#',
            'fontFamily': 'Rubik',
            'axisPointer': {'type': 'shadow'},
            'type': 'category',
            'nameLocation': 'middle',
            'nameGap': 35,
        }],
        'yAxis': [
            {
                'type': 'value',
                'fontFamily': 'Rubik',
                'axisLabel': {
                    'formatter': '{value}'+bar_suffix
                },
                'nameLocation': 'middle',
                'nameGap': 24+7*(longest_value_bars+int(len(bar_suffix))),
                'alignTicks': True,
                'axisLine': {
                    'show': True,
                    'lineStyle': {
                        'color': ShimokuPalette.CHART_C1.value
                    }
                },
                'splitNumber': 5,
                'interval': bar_interval,
                'max': max_bars,
            },
            {
                'type': 'value',
                'fontFamily': 'Rubik',
                'axisLabel': {
                    'formatter': '{value}'+line_suffix
                },
                'nameLocation': 'middle',
                'nameGap': 24+7*(longest_value_lines+int(len(line_suffix))),
                'axisLine': {
                    'show': True,
                    'lineStyle': {
                        'color': f'var(--chart-C{len(bar_names)+2})'
                    }
                },
                'splitNumber': 5,
                'interval': line_interval,
                'max': max_lines,
            }
        ],
    })
    series_options = get_common_series_options()

    series = []
    for index, name in enumerate(bar_names):
        sop = deepcopy(series_options)
        sop.update({
            'name': name,
            'type': 'bar',
            'itemStyle': {
                'color': f'var(--chart-C{index+1})',
                'borderRadius': [9, 9, 0, 0]
            },
            'emphasis':
                {
                    'itemStyle': {
                        'color': f'var(--chart-C{index+1})',
                        'borderRadius': [9, 9, 0, 0]
                    },
                    'lineStyle': {
                        'color': f'var(--chart-C{index+1})',
                    }
                },
            'smooth': True,
        })
        series.append(sop)

    for index, name in enumerate(line_names):
        sop = deepcopy(series_options)
        sop.update({
            'name': name,
            'type': 'line',
            'yAxisIndex': 1,
            'itemStyle': {
                'color': f'var(--chart-C{len(bar_names) + index + 1})',
                'borderRadius': [9, 9, 0, 0]
            },
            'emphasis':
                {
                    'itemStyle': {
                        'color': f'var(--chart-C{len(bar_names) + index + 1})',
                        'borderRadius': [9, 9, 0, 0]
                    },
                    'lineStyle': {
                        'color': f'var(--chart-C{len(bar_names) + index + 1})',
                    }
                },
            'smooth': True,
        })
        series.append(sop)

    await self._create_trend_chart(
        echart_options=common_options, series_options=series,
        axes=x, values=bar_names+line_names, option_modifications=option_modifications,
        data_mapping_to_tuples=data_mappings_to_tuples,
        order=order, x_axis_names=x_axis_name,
        y_axis_names=[bar_axis_name, line_axis_name], variant=variant,
        title=title, rows_size=rows_size, cols_size=cols_size, padding=padding,
    )



from typing import Union, Optional, TypedDict, TYPE_CHECKING
from shimoku.api.base_resource import Resource

if TYPE_CHECKING:
    from shimoku.api.resources.business import Business
    from shimoku.api.resources.dashboard import Dashboard
    from shimoku.api.resources.app import App
    from shimoku.api.user_access_classes.businesses_layer import WorkspacesLayer
    from shimoku.api.user_access_classes.apps_layer import MenuPathsLayer
    from shimoku.api.user_access_classes.dashboards_layer import BoardsLayer
import logging
from shimoku.execution_logger import log_error
logger = logging.getLogger(__name__)

VALID_PERMISSIONS = ['READ', 'WRITE']
VALID_RESOURCES = ['DATA', 'DATA_EXECUTION', 'USER_MANAGEMENT', 'BUSINESS_INFO']
VALID_TARGETS = ['GROUP', 'USER']


class Role(Resource):

    resource_type = 'role'
    alias_field = 'role'
    plural = 'roles'

    class RoleParams(TypedDict):
        permission: str
        resource: str
        target: str
        role: str

    def __init__(self, parent: Union['Business', 'Dashboard', 'App'], uuid: Optional[str] = None,
                 alias: Optional[str] = None, db_resource: Optional[dict] = None):
        params: Role.RoleParams = dict(
            permission='READ',
            resource='BUSINESS_INFO',
            target='GROUP',
            role=alias if alias else '',
        )
        super().__init__(parent=parent, uuid=uuid, check_params_before_creation=['role'], params=params,
                         db_resource=db_resource)

    async def delete(self):
        return await self._base_resource.delete()

    async def update(self):
        return await self._base_resource.update()

    def set_params(self, permission: Optional[str] = None, resource: Optional[str] = None,
                   target: Optional[str] = None, role: Optional[str] = None):

        if permission:
            if permission and permission not in VALID_PERMISSIONS:
                log_error(logger, f'{permission} is not a valid value for permission, '
                                  f'the valid values are: {VALID_PERMISSIONS}', ValueError)
            self._base_resource.params['permission'] = permission
        if resource:
            if resource not in VALID_RESOURCES:
                log_error(logger, f'{resource} is not a valid value for resource, '
                                  f'the valid values are: {VALID_RESOURCES}', ValueError)
            self._base_resource.params['resource'] = resource
        if target:
            if target and target not in VALID_TARGETS:
                log_error(logger, f'{target} is not a valid value for target, '
                                  f'the valid values are: {VALID_TARGETS}', ValueError)
            self._base_resource.params['target'] = target
        if role:
            self._base_resource.params['role'] = role


# Parent class methods
async def create_role(
    self: Union['Business', 'Dashboard', 'App'], role: str, permission: str = 'READ',
    resource: str = 'BUSINESS_INFO', target: str = 'GROUP'
) -> Role:
    return await self._base_resource.create_child(Role, alias=role, permission=permission,
                                                  resource=resource, target=target)


async def get_role(
    self: Union['Business', 'Dashboard', 'App'], uuid: Optional[str] = None, role: Optional[str] = None
) -> Optional[Role]:
    return await self._base_resource.get_child(Role, uuid, role)


async def get_roles(
    self: Union['Business', 'Dashboard', 'App']
) -> list[Role]:
    return await self._base_resource.get_children(Role)


async def delete_role(
    self: Union['Business', 'Dashboard', 'App'], uuid: Optional[str] = None, role: Optional[str] = None
) -> Role:
    return await self._base_resource.delete_child(Role, uuid, role)


# User level methods
async def user_create_role(
    self: Union['WorkspacesLayer', 'MenuPathsLayer', 'BoardsLayer'],
    uuid: Optional[str] = None, name: Optional[str] = None,
    resource: Optional[str] = None, role_name: Optional[str] = None,
    permission: Optional[str] = None, target: Optional[str] = None
) -> dict:
    """
    Create a new role at the resource level
    :param name: name of the resource
    :param uuid: uuid of the resource
    :param resource: resource of the role
    :param role_name: name of the role
    :param permission: permission of the role
    :param target: target of the role
    """
    resource_obj = await self._get_for_roles(uuid, name)

    return (await resource_obj.create_role(
                role=role_name, permission=permission, resource=resource, target=target)
            ).cascade_to_dict()


async def user_delete_role(
    self: Union['WorkspacesLayer', 'MenuPathsLayer', 'BoardsLayer'],
    uuid: Optional[str] = None, name: Optional[str] = None,
    role_id: Optional[str] = None, role_name: Optional[str] = None
):
    """
    Delete the role at the resource level
    :param name: name of the resource
    :param uuid: uuid of the resource
    :param role_id: id of the role
    :param role_name: name of the role
    """
    resource = await self._get_for_roles(uuid, name)

    if resource:
        await resource.delete_role(uuid=role_id, role=role_name)


async def user_get_role(
    self: Union['WorkspacesLayer', 'MenuPathsLayer', 'BoardsLayer'],
    uuid: Optional[str] = None, name: Optional[str] = None,
    role_name: Optional[str] = None
) -> list[dict]:
    """
    Get the role at the resource level
    :param name: name of the resource
    :param uuid: uuid of the resource
    :param role_name: name of the role
    """
    resource = await self._get_for_roles(uuid, name)

    if resource:
        role = await resource.get_role(role=role_name)
        if role:
            return role.cascade_to_dict()
        logger.info(f'Role {role_name} not found in {str(resource)}')

    return []


async def user_get_roles(
    self: Union['WorkspacesLayer', 'MenuPathsLayer', 'BoardsLayer'],
    uuid: Optional[str] = None, name: Optional[str] = None
) -> list[dict]:
    """
    Get the roles at the resource level
    :param name: name of the resource
    :param uuid: uuid of the resource
    """
    resource = await self._get_for_roles(uuid, name)

    if resource:
        return [role.cascade_to_dict() for role in await resource.get_roles()]

    return []



from typing import Optional

import os

from shimoku.ai.ai_layer import (
    get_model_metadata, check_and_get_model, get_output_file_metadata, get_activity_template,
    get_activity_from_template, get_runs_message
)

from shimoku.api.resources.file import File
from shimoku.api.resources.app import App
from shimoku.api.resources.activity import Activity
from shimoku.api.resources.activity_template import ActivityTemplate

from shimoku.cli import CLIParser, CLIFuncParam
from shimoku.cli.utils import list_filtering_and_display, display_list, save_as_file
from shimoku.cli.cloud.cascade_get_resources import InitOptions, ResourceGetter, Universe

import logging
logger = logging.getLogger(__name__)

def add_list_parser(parser: Optional[CLIParser] = None):
    """
    Function to add the list parser to a parser
    :param parser: Parser to add the listing parser to
    :return: Listing parser
    """
    params = {
        'name': 'list',
        'description': 'Commands to list resources'
    }
    if parser:
        list_parser = parser.add_command(**params)
    else:
        list_parser = CLIParser(**params)
        
    common_args = [
        CLIFuncParam(name='local-port', arg_type=int, arg_help='Local port to use', mandatory=False),
        CLIFuncParam(name='environment', arg_type=str, arg_help='Environment to use', mandatory=False),
        CLIFuncParam(name='access-token', arg_type=str, arg_help='Access token to use', mandatory=False),
        CLIFuncParam(name='universe-id', arg_type=str, arg_help='Universe ID to use', mandatory=False),
        CLIFuncParam(name='count', arg_help='Flag to show the count of items', action='store_true', mandatory=False),
        CLIFuncParam(name='prefix', arg_type=str, arg_help='Prefix to filter the items', mandatory=False),
        CLIFuncParam(name='contains', arg_type=str, arg_help='String to filter the items', mandatory=False),
        CLIFuncParam(name='filter-field', arg_type=str, arg_help='Field to filter the items', mandatory=False),
        CLIFuncParam(name='sort-field', arg_type=str, arg_help='Field to sort the items', mandatory=False),
        CLIFuncParam(name='equal-to', arg_type=str, arg_help='String to match the items', mandatory=False),
        CLIFuncParam(name='not-equal-to', arg_type=str, arg_help='String to not match the items', mandatory=False),
        CLIFuncParam(name='case-sensitive', arg_help='Flag to set the case sensitivity of the string filters',
                     action='store_true', mandatory=False),
    ]

    module_functions = [
        universe_api_keys,
        activity_templates,
        workspaces,
        boards,
        menu_paths,
        components,
        sub_paths,
        data_sets,
        data,
        activities,
        runs,
        run_logs,
        files,
        ai_input_files,
        ai_functions,
        ai_function_parameters,
        ai_models,
        ai_execution_logs,
        ai_output_files,
        roles,
    ]

    for func in module_functions:
        list_parser.decor_add_command(common_args=common_args)(func)

    return list_parser


async def universe_api_keys(**kwargs):
    """ List the universe api keys in the universe """
    universe: Universe = await ResourceGetter(InitOptions(**kwargs)).get_universe()
    list_filtering_and_display(
        await universe.get_universe_api_keys(),
        table_title=f"Universe api keys from Universe {universe['id']}",
        fields=['id', 'description', 'enabled'],
        **kwargs
    )


async def workspaces(
    show_theme: bool = CLIFuncParam(default=False, action='store_true', mandatory=False),
    **kwargs,
):
    """ List the workspaces in the universe
    :param show_theme: Flag to show the theme
    """
    resource_getter = ResourceGetter(InitOptions(**kwargs))
    universes_layer = await resource_getter.get_universes_layer()
    universe = await resource_getter.get_universe()
    if kwargs['filter_field'] is None:
        kwargs['filter_field'] = 'name'
    if not kwargs['sort_field'] and kwargs['filter_field'] == 'theme':
        kwargs['sort_field'] = 'name'
    list_filtering_and_display(
        await universes_layer.get_universe_workspaces(universe['id']),
        table_title=f"Workspaces from Universe {universe['id']}",
        fields=resource_getter.get_business_fields_to_show(show_theme),
        table_lines=False,
        **kwargs
    )


async def activity_templates(
    show_input_settings: bool = CLIFuncParam(default=False, action='store_true', mandatory=False),
    **kwargs,
):
    """ List the activity templates in the universe
    :param show_input_settings: Flag to show the input settings
    """
    resource_getter = ResourceGetter(InitOptions(**kwargs))
    universes_layer = await resource_getter.get_universes_layer()
    universe = await resource_getter.get_universe()
    list_filtering_and_display(
        await universes_layer.get_universe_activity_templates(universe['id']),
        table_title=f"Activity templates from Universe {universe['id']}",
        fields=resource_getter.get_activity_template_fields_to_show(show_input_settings),
        **kwargs
    )


async def boards(
    workspace_id: Optional[str],
    show_public_permission: bool = CLIFuncParam(default=False, action='store_true', mandatory=False),
    **kwargs,
):
    """ List the boards in the workspace
    :param workspace_id: UUID of the workspace to use
    :param show_public_permission: Flag to show the public permission
    """
    resource_getter = ResourceGetter(InitOptions(workspace_id=workspace_id, **kwargs))
    businesses_layer = await resource_getter.get_businesses_layer()
    business = await resource_getter.get_business()
    board_items = await businesses_layer.get_workspace_boards(business['id'])
    fields = resource_getter.get_dashboard_fields_to_show(show_public_permission)
    if kwargs['filter_field'] is None:
        kwargs['filter_field'] = 'name'
    if not kwargs['sort_field']:
        kwargs['sort_field'] = 'order'
    list_filtering_and_display(
        board_items,
        table_title=f"Boards from Workspace {str(business)}",
        fields=fields,
        **kwargs
    )


async def menu_paths(
    board: Optional[str],
    workspace_id: Optional[str],
    **kwargs
):
    """ List the menu paths in the board
    :param board: Board id or name to use
    :param workspace_id: UUID of the workspace to use
    """
    resource_getter = ResourceGetter(InitOptions(workspace_id=workspace_id, board=board, **kwargs))
    business_layer = await resource_getter.get_businesses_layer()
    business = await resource_getter.get_business()
    board_obj = None
    if board:
        board_obj = await resource_getter.get_dashboard()
    fields = resource_getter.get_app_fields_to_show()
    if kwargs['filter_field'] is None:
        kwargs['filter_field'] = 'name'
    if not kwargs['sort_field']:
        kwargs['sort_field'] = 'order'
    menu_path_objs = await business_layer.get_workspace_menu_paths(business['id'])
    if board_obj:
        board_menu_path_ids = await board_obj.list_app_ids()
        menu_path_objs = [menu_path for menu_path in menu_path_objs if menu_path['id'] in board_menu_path_ids]
    list_filtering_and_display(
        menu_path_objs,
        table_title=f"Menu paths from Workspace {str(business)}"
                    if not board_obj else
                    f"Menu paths from Board {str(board_obj)}",
        fields=fields,
        **kwargs
    )


async def data_sets(
    workspace_id: Optional[str],
    menu_path: str = CLIFuncParam(prompt=True),
    **kwargs
):
    """ List the data sets in the menu path
    :param workspace_id: UUID of the workspace to use
    :param menu_path: Menu path id or name to use
    """
    resource_getter = ResourceGetter(InitOptions(workspace_id=workspace_id, menu_path=menu_path, **kwargs))
    apps_layer = await resource_getter.get_apps_layer()
    app = await resource_getter.get_app()
    fields = resource_getter.get_data_set_fields_to_show()
    if kwargs['filter_field'] is None:
        kwargs['filter_field'] = 'name'
    list_filtering_and_display(
        await apps_layer.get_menu_path_data_sets(app['id']),
        table_title=f"Data sets from Menu path {str(app)}",
        table_lines=False,
        fields=fields,
        **kwargs
    )


async def components(
    workspace_id: Optional[str],
    sub_path: Optional[str],
    menu_path: str = CLIFuncParam(prompt=True),
    show_all_fields: bool = CLIFuncParam(default=False, action='store_true', mandatory=False),
    **kwargs
):
    """ List the components in the menu path
    :param workspace_id: UUID of the workspace to use
    :param menu_path: Menu path id or name to use
    :param show_all_fields: Flag to show all the fields
    :param sub_path: Sub path to filter the components
    """
    resource_getter = ResourceGetter(InitOptions(
        workspace_id=workspace_id,
        menu_path=menu_path,
        **kwargs
    ))
    apps_layer = await resource_getter.get_apps_layer()
    app = await resource_getter.get_app()
    fields = resource_getter.get_report_fields_to_show(show_all_fields=show_all_fields)
    reports = await apps_layer.get_menu_path_components(app['id'])
    if sub_path:
        reports = [report for report in reports if report['path'] == sub_path]
        fields.pop(fields.index('path'))
        fields.pop(fields.index('pathOrder'))
    if kwargs['filter_field'] is None:
        kwargs['filter_field'] = 'reportType'
    if not kwargs['sort_field']:
        kwargs['sort_field'] = 'pathOrder'
    list_filtering_and_display(
        reports,
        table_title=f"Components from menu path ({str(app)+(f'/{sub_path}' if sub_path else '')})",
        fields=fields,
        table_lines=show_all_fields,
        **kwargs
    )


async def sub_paths(
    workspace_id: Optional[str],
    menu_path: str = CLIFuncParam(prompt=True),
    **kwargs
):
    """ List the sub paths in the menu path
    :param workspace_id: UUID of the workspace to use
    :param menu_path: Menu path id or name to use
    """
    resource_getter = ResourceGetter(InitOptions(workspace_id=workspace_id, menu_path=menu_path, **kwargs))
    apps_layer = await resource_getter.get_apps_layer()
    app = await resource_getter.get_app()
    display_list(
        f'Sub paths from Menu path {str(app)}',
        await apps_layer.get_menu_path_sub_paths(app['id']),
    )


async def data(
    workspace_id: Optional[str],
    menu_path: str = CLIFuncParam(prompt=True),
    data_set: str = CLIFuncParam(prompt=True),
    limit: int = CLIFuncParam(default=100, mandatory=False),
    **kwargs
):
    """ List the data in the data set
    :param workspace_id: UUID of the workspace to use
    :param menu_path: Menu path id or name to use
    :param data_set: Data set id or name to use
    :param limit: Limit the number of results returned, negative or zero to show all
    """
    resource_getter = ResourceGetter(InitOptions(
        workspace_id=workspace_id,
        menu_path=menu_path,
        data_set=data_set,
        **kwargs
    ))
    data_sets_layer = await resource_getter.get_data_sets_layer()
    data_set = await resource_getter.get_data_set()
    data = await data_sets_layer.get_data_from_data_set(data_set['id'], limit=limit if limit > 0 else None)
    show_custom_field = 'customField1' in data[0] if data else False
    fields = resource_getter.get_data_fields_to_show(show_custom_field)
    list_filtering_and_display(
        data,
        table_title=f"Data from Data set ({str(data_set)})",
        fields=fields,
        table_lines=False,
        **kwargs
    )


async def activities(
    workspace_id: Optional[str],
    menu_path: str = CLIFuncParam(prompt=True),
    show_templates: bool = CLIFuncParam(default=False, action='store_true', mandatory=False),
    **kwargs
):
    """ List the activities in the menu path
    :param workspace_id: UUID of the workspace to use
    :param menu_path: Menu path id or name to use
    :param show_templates: Flag to show the templates
    """
    resource_getter = ResourceGetter(InitOptions(workspace_id=workspace_id, menu_path=menu_path, **kwargs))
    apps_layer = await resource_getter.get_apps_layer()
    app = await resource_getter.get_app()
    fields = resource_getter.get_activity_fields_to_show(show_templates)
    if kwargs['filter_field'] is None:
        kwargs['filter_field'] = 'name'
    list_filtering_and_display(
        await apps_layer.get_menu_path_activities(app['id']),
        table_title=f"Activities from Menu path {str(app)}",
        fields=fields,
        **kwargs
    )


async def runs(
    workspace_id: Optional[str],
    menu_path: str = CLIFuncParam(prompt=True),
    activity: str = CLIFuncParam(prompt=True),
    **kwargs
):
    """ List the runs of an activity
    :param workspace_id: UUID of the workspace to use
    :param menu_path: Menu path id or name to use
    :param activity: Activity id or name to use
    """
    resource_getter = ResourceGetter(InitOptions(
        workspace_id=workspace_id,
        menu_path=menu_path,
        activity=activity,
        **kwargs
    ))
    activity = await resource_getter.get_activity()
    runs_objs = await activity.get_runs()
    list_filtering_and_display(
        [run.cascade_to_dict() for run in runs_objs],
        table_title=f"Runs from Activity {str(activity)}",
        **kwargs
    )


async def run_logs(
    workspace_id: Optional[str],
    menu_path: str = CLIFuncParam(prompt=True),
    activity: str = CLIFuncParam(prompt=True),
    run_id: str = CLIFuncParam(prompt=True),
    **kwargs
):
    """ List the logs of a run
    :param workspace_id: UUID of the workspace to use
    :param menu_path: Menu path id or name to use
    :param activity: Activity id or name to use
    :param run_id: Run id to use
    """
    resource_getter = ResourceGetter(InitOptions(
        workspace_id=workspace_id,
        menu_path=menu_path,
        activity=activity,
        run_id=run_id,
        **kwargs
    ))
    activity = await resource_getter.get_activity()
    run = await activity.get_run(run_id)
    list_filtering_and_display(
        run['logs'],
        table_title=f"Logs from Run {run_id}",
        **kwargs
    )


async def files(
    workspace_id: Optional[str],
    menu_path: str = CLIFuncParam(prompt=True),
    with_shimoku_generated: bool = CLIFuncParam(default=False, action='store_true', mandatory=False),
    **kwargs
):
    """ List the files in the menu path
    :param workspace_id: UUID of the workspace to use
    :param menu_path: Menu path id or name to use
    :param with_shimoku_generated: Flag to show the shimoku generated files
    """
    resource_getter = ResourceGetter(InitOptions(workspace_id=workspace_id, menu_path=menu_path, **kwargs))
    apps_layer = await resource_getter.get_apps_layer()
    app = await resource_getter.get_app()
    fields = resource_getter.get_file_fields_to_show()
    if kwargs['filter_field'] is None:
        kwargs['filter_field'] = 'name'
    list_filtering_and_display(
        await apps_layer.get_menu_path_files(app['id'], with_shimoku_generated=with_shimoku_generated),
        table_title=f"Files from Menu path {str(app)}",
        fields=fields,
        **kwargs
    )


async def roles(
    workspace_id: Optional[str],
    menu_path: str = CLIFuncParam(group='menu_path_or_board', mandatory=False),
    board: str = CLIFuncParam(group='menu_path_or_board', mandatory=False),
    **kwargs
):
    """ List the roles in the menu path, board or workspace. If menu path is specified, it will list the roles in the
    menu path. If board is specified, it will list the roles in the board. If none is specified, it will list the roles
    in the workspace. Don't specify both menu path and board.
    :param workspace_id: UUID of the workspace to use
    :param menu_path: Menu path id or name to use
    :param board: Board id or name to use
    """
    resource_getter = ResourceGetter(InitOptions(
        workspace_id=workspace_id,
        menu_path=menu_path,
        board=board,
        **kwargs
    ))

    if menu_path:
        app = await resource_getter.get_app()
        title = f"Roles from Menu path {str(app)}"
    elif board:
        dashboard = await resource_getter.get_dashboard()
        title = f"Roles from Board {str(dashboard)}"
    else:
        business = await resource_getter.get_business()
        title = f"Roles from Workspace {str(business)}"

    list_filtering_and_display(
        await resource_getter.get_role_dicts(),
        table_title=title,
        **kwargs
    )


# AI functions

async def ai_input_files(
    workspace_id: Optional[str],
    menu_path: str = CLIFuncParam(prompt=True),
    **kwargs
):
    """ List the AI input files in the menu path
    :param workspace_id: UUID of the workspace to use
    :param menu_path: Menu path id or name to use
    """
    ai_layer = await ResourceGetter(
        InitOptions(
            workspace_id=workspace_id,
            menu_path=menu_path,
            **kwargs
        )
    ).get_ai_layer()
    message = [
        "",
        "///////////////////////////////////////////////////",
        "///////////////// Available Files /////////////////",
    ]
    input_files = await ai_layer.get_available_input_files()
    for file in input_files:
        metadata = {k: v for k, v in file.items() if k not in ['file_name']}
        message.extend([
            '',
            f" \033[1m- File name:\033[0m {file['file_name']}",

        ])
        if metadata:
            message.extend([
                f"   \033[1mMetadata:\033[0m",
                *[f"     \033[1m- {k}:\033[0m {v}" for k, v in metadata.items()]
            ])

    message.extend(["", "///////////////////////////////////////////////////", ""])
    print('\n'.join(message))


def add_input_params_to_message(params: dict[str, dict]):
    """ Add the parameters to a message
    :param params: Parameters to add
    """
    message = []
    for param_name, param in params.items():
        message.append(f"\033[1m- {param_name}"
                       f"{' (Optional)' if not param['mandatory'] else ''}:\033[0m {param['datatype']}")
        if param['description']:
            message.append(f"  {param['description']}")
    return message


def add_args_and_files_to_message(input_settings: dict[str, dict]) -> list[str]:
    """ Add the args and files to a message
    :param input_settings: Input settings of the ai_function
    """
    message = []
    input_files = {key: param for key, param in input_settings.items()
                   if isinstance(param, dict) and param['datatype'] == 'file'}
    input_args = {key: param for key, param in input_settings.items()
                  if isinstance(param, dict) and param['datatype'] != 'file'}
    message.append(f"--- args ---")
    message.extend(["  " + line for line in add_input_params_to_message(input_args)])
    message.append(f"--- files ---")
    message.extend(["  " + line for line in add_input_params_to_message(input_files)])
    return message


async def ai_functions(
    show_input_parameters: bool = CLIFuncParam(default=False, action='store_true', mandatory=False),
    **kwargs
):
    """ Show the available ai_functions
    :param show_input_parameters: Show the input parameters of each ai_function
    """
    universe = await ResourceGetter(InitOptions(**kwargs)).get_universe()
    message = [
        "",
        "///////////////////////////////////////////////////",
        "/////////////// Available ai_functions ///////////////",
    ]
    templates = await universe.get_activity_templates()
    while len(templates) > 0:
        template = templates.pop(0)
        if not template['enabled']:
            continue
        versions = [template['version']]
        while len(templates) > 0 and templates[0]['name'] == template['name']:
            template = templates.pop(0)
            versions.append(template['version'])
        message.append("")
        message.append(f" \033[1m- AI function:\033[0m {template['name']} (v{', v'.join(versions)})")
        message.append(
            f"   \033[1mDescription:\033[0m {template['description']} "
            f"(wait time between runs: >{template['minRunInterval']}s)"
        )
        if show_input_parameters:
            message.append(f"   \033[1mInput parameters:\033[0m")
            message.extend(["     " + line
                            for line in add_args_and_files_to_message(template['inputSettings'])])

    message.extend(["", "///////////////////////////////////////////////////", ""])
    print('\n'.join(message))


async def ai_function_parameters(
    ai_function: str = CLIFuncParam(prompt=True),
    **kwargs
):
    """ Show the parameters of a ai_function
    :param ai_function: Name of the ai_function
    """
    universe = await ResourceGetter(InitOptions(**kwargs)).get_universe()
    activity_templates = await universe.get_activity_templates()
    template = [template for template in activity_templates if template['name'] == ai_function][-1]
    message = [
        "",
        "///////////////////////////////////////////////////",
        f" {ai_function} parameters ".center(51, '/'),
        "",
        f"  \033[1mDescription:\033[0m {template['description']} "
        f"(wait time between runs: >{template['minRunInterval']}s)",
        f"  \033[1mInput parameters:\033[0m"
    ]
    message.extend(["    " + line for line in add_args_and_files_to_message(template['inputSettings'])])

    message.extend(["", "///////////////////////////////////////////////////", ""])
    print('\n'.join(message))


async def ai_models(
    workspace_id: Optional[str],
    menu_path: str = CLIFuncParam(prompt=True),
    **kwargs
):
    """ Show the available models
    :param workspace_id: UUID of the workspace to use
    :param menu_path: Menu path id or name to use
    """
    message = [
        "",
        "///////////////////////////////////////////////////",
        "//////////////// Available models /////////////////",
    ]
    resource_getter = ResourceGetter(InitOptions(workspace_id=workspace_id, menu_path=menu_path, **kwargs))
    app = await resource_getter.get_app()
    app_files: list[File] = await app.get_files()
    model_files = [file for file in app_files if 'shimoku_generated' in file['tags'] and 'ai_model' in file['tags']]
    model_files = sorted(model_files, key=lambda file: file['metadata']['model_name'])
    for file in model_files:
        message.extend([
            '',
            f" \033[1m- Model name:\033[0m {file['metadata']['model_name']}",
            f"   \033[1mMetadata:\033[0m"
        ])
        for key, value in get_model_metadata(file).items():
            if key in ['model_name', 'creator_ai_function_version']:
                continue
            message.append(f"     \033[1m- {key}:\033[0m {value}")

    message.extend(["", "///////////////////////////////////////////////////", ""])
    print('\n'.join(message))


async def ai_execution_logs(
    workspace_id: Optional[str],
    menu_path: str = CLIFuncParam(prompt=True),
    ai_function: str = CLIFuncParam(prompt=True),
    version: str = CLIFuncParam(mandatory=False),
    run_id: str = CLIFuncParam(group='filtering', mandatory=False),
    model_name: str = CLIFuncParam(group='filtering', mandatory=False),
    how_many: int = CLIFuncParam(default=1, mandatory=False),
    **kwargs
):
    """ Show the logs of the executions of a ai_function
    :param workspace_id: UUID of the workspace to use
    :param menu_path: Menu path id or name to use
    :param ai_function: Name of the ai_function to execute
    :param version: Version of the ai_function to execute
    :param run_id: Id of the run to get the logs from
    :param model_name: Name of the model to get the logs from
    :param how_many: Number of executions to get
    """

    async def show_last_execution_logs_by_model(
            _app: App,
    ):
        app_files: list[File] = await _app.get_files()
        model_file: File = await check_and_get_model(_app, model_name)
        model_metadata: dict = get_model_metadata(model_file)
        runs: list[Activity.Run] = []
        input_settings_by_run_id: dict[str, dict] = {}
        for file in app_files:
            if 'shimoku_generated' not in file['tags'] or 'ai_output_file' not in file['tags']:
                continue
            output_file_metadata: dict = get_output_file_metadata(file)
            output_file_model_name = output_file_metadata.pop('model_name') if model_metadata else None
            if output_file_model_name == model_name:
                activity_template: ActivityTemplate = await get_activity_template(
                    _app.parent.parent,
                    output_file_metadata['creator_ai_function'],
                    output_file_metadata['creator_ai_function_version']
                )
                activity: Activity = await get_activity_from_template(
                    _app, activity_template, create_if_not_exists=False
                )
                run_id = output_file_metadata['run_id']
                input_settings = activity_template['inputSettings']
                runs.append(await activity.get_run(run_id))
                input_settings_by_run_id[run_id] = input_settings
        runs = (await Activity.sort_runs_by_log_time(runs))[-how_many:]
        message = [
            '',
            '///////////////////////////////////////////////////',
            f' LOGS OF EXECUTIONS BY {model_name.upper()} '.center(51, '/')
        ]
        input_settings_list = [input_settings_by_run_id[run['id']] for run in runs]
        message.extend(await get_runs_message(runs, input_settings_list))
        print('\n'.join(message))

    async def show_last_execution_logs_by_ai_function(
            _app: App,
    ):
        activity_template: ActivityTemplate = await get_activity_template(_app.parent.parent, ai_function, version)
        input_settings = activity_template['inputSettings']
        activity: Activity = await get_activity_from_template(_app, activity_template, create_if_not_exists=False)
        runs: list[Activity.Run] = await activity.get_runs(how_many)
        if run_id:
            runs = [run for run in runs if run['id'] == run_id]
        message = [
            '',
            '///////////////////////////////////////////////////',
            f' LOGS OF {ai_function.upper()} '.center(51, '/')
        ]
        message.extend(await get_runs_message(runs, input_settings))
        print('\n'.join(message))

    resource_getter = ResourceGetter(InitOptions(workspace_id=workspace_id, menu_path=menu_path, **kwargs))
    app = await resource_getter.get_app()
    if model_name:
        await show_last_execution_logs_by_model(app)
    else:
        await show_last_execution_logs_by_ai_function(app)


async def ai_output_files(
    workspace_id: Optional[str],
    menu_path: str = CLIFuncParam(prompt=True),
    ai_function: str = CLIFuncParam(mandatory=False),
    version: str = CLIFuncParam(mandatory=False),
    model_name: str = CLIFuncParam(mandatory=False),
    file_name: str = CLIFuncParam(mandatory=False),
    save_to_path: str = CLIFuncParam(mandatory=False),
    **kwargs
):
    """ Show the output file of the execution of a ai_function
    :param workspace_id: UUID of the workspace to use
    :param menu_path: Menu path id or name to use
    :param ai_function: Name of the ai_function to execute
    :param version: Version of the ai_function to execute
    :param model_name: Name of the model to get the logs from
    :param file_name: Name of the file to get
    :param save_to_path: Path to store the file objects, leave empty to not download the files content
    """
    resource_getter = ResourceGetter(InitOptions(workspace_id=workspace_id, menu_path=menu_path, **kwargs))
    ai_layer = await resource_getter.get_ai_layer()
    files_dict = []
    if ai_function:
        files_dict.extend(await ai_layer.get_output_files_by_ai_function(
            ai_function, version, file_name, get_objects=str(save_to_path) != ''
        ))
    elif model_name:
        files_dict.extend(await ai_layer.get_output_files_by_model(
            model_name, file_name, get_objects=str(save_to_path) != ''
        ))
    else:
        raise ValueError('You must specify either the ai_function or the model_name')
    if save_to_path:
        for run in files_dict:
            file_dir = os.path.join(save_to_path, run['id'])
            for file_name, file_obj in run['files']:
                save_as_file(logger, file_dir, file_obj, file_name)
    list_filtering_and_display(
        files_dict,
        table_title=f"Output files from {ai_function or model_name}",
        **kwargs
    )


if __name__ == "__main__":
    add_list_parser().parse_args()
